"""
Atomic Correct potential file
"""
import sys
import click
from loguru import logger
from minushalf.softwares import VaspFactory
from minushalf.data import Softwares
from minushalf.utils import (welcome_message, end_message, Vtotal,
                             AtomicPotential, parse_cut)


@click.command()
@click.option('--quiet', default=False, is_flag=True)
@click.option('-b',
              '--base_potfile_path',
              type=click.Path(),
              help="""Path to the folder containing the potential file""")
@click.option('-v',
              '--vtotal_path',
              type=click.Path(exists=True),
              default="VTOTAL.ae",
              show_default=True,
              help="""Path relative to the potential
              file generated by the atomic program for the
              atom with all electrons.""")
@click.option('-o',
              '--vtotal_occupied_path',
              type=click.Path(exists=True),
              default="VTOTAL_OCC",
              show_default=True,
              help="""Path relative to the potential
              file generated by the atomic program for the
              occupied atom.""")
@click.option(
    '-s',
    '--software',
    type=click.Choice(Softwares.to_list(), case_sensitive=False),
    default=Softwares.vasp.value,
    show_default=True,
    help="""Specifies the software used to make ab initio calculations.""")
@click.option('-c',
              '--correction',
              type=click.Choice(['VALENCE', 'CONDUCTION'],
                                case_sensitive=False),
              default="VALENCE",
              show_default=True,
              help="""Indicates whether the correction should be
                      made in the valence band or the conduction band.""")
@click.option('-C',
              '--cut',
              type=str,
              nargs=1,
              default="2.0",
              show_default=True,
              help="""
              distance value used to cut the potential generated artificially
              by fractional atomic occupation , it can be passed in two ways:

              unique value : float or integer. Ex: 1.0

              \b
              range:  begin(float|integer):pass(float|integer):end(float|integer). Ex: 1.0:0.1:2.0

              """)
@click.option(
    '-a',
    '--amplitude',
    type=float,
    default=1.0,
    show_default=True,
    nargs=1,
    help=
    """Scaling factor to be used to correct the artificially generated potential.
               In the vast majority of cases, the amplitude value is 1.0. However, there are some
               special cases where this value needs to be adjusted. Therefore, we recommend that
                you do not change this value unless you know exactly what you are doing"""
)
def correct_potfile(
    quiet: bool,
    base_potfile_path: str,
    vtotal_path: str,
    vtotal_occupied_path: str,
    software: str,
    correction: str,
    cut: str,
    amplitude: float,
) -> None:
    """Generate the occupied atomic potential file used for ab initio calculations.

    Requires:

        VTOTAL.ae: potential of the atom with all electrons

        VTOTAL_OCC: potential of the occupied atom

        INP_OCC: Input file for the run-atomic command of the occupied atom

        The command also needs the potential files used by the chosen software:

            VASP: POTCAR (This name can't be changed)


    Generates:

        POTFILEcut${CUT_VALUE} (If amplitude is equal to 1.0)

        POTFILEcut${CUT_VALUE}A${AMPLITUDE_VALUE} (If amplitude is different from 1.0)

    """
    welcome_message("minushalf")

    if quiet:
        logger.remove()
        logger.add(sys.stdout, level="ERROR")

    softwares = {"VASP": VaspFactory()}
    factory = softwares[software.upper()]

    vtotal = Vtotal.from_file(vtotal_path)
    vtotal_occ = Vtotal.from_file(vtotal_occupied_path)
    potential_file = factory.get_potential_class(base_path=base_potfile_path)

    atomic_potential = AtomicPotential(vtotal, vtotal_occ, potential_file)

    cut_numbers = parse_cut(cut)

    is_conduction = False
    if correction.upper() == "CONDUCTION":
        is_conduction = True

    for new_cut in cut_numbers:
        logger.info("Correcting POTFILE for cut = {:.2f} a.u".format(new_cut))
        new_potential = atomic_potential.correct_potential(
            new_cut, amplitude, is_conduction)
        atomic_potential.correct_file(new_potential, new_cut, amplitude,
                                      is_conduction)

    end_message()



"""
Init file for interfaces
"""
from .software_abstract_factory import SoftwaresAbstractFactory
from .potential_file import PotentialFile
from .band_projection_file import BandProjectionFile
from .runner import Runner
from .correction import Correction



"""
List calculation code options for the INP file
"""
from enum import Enum, unique


@unique
class CalculationCode(Enum):
    """
    Enum type for the calculation code
    """

    ae = "ae"  # All electrons

    def __str__(self):
        return str(self.name)

    @staticmethod
    def to_list():
        """
        Generate list of available calculation codes
        """
        return list(map(lambda element: element.value, CalculationCode))



"""
Reads procar file, an output of
VASP software
"""
import re
from itertools import islice
from loguru import logger
from minushalf.interfaces import BandProjectionFile


class Procar(BandProjectionFile):
    """
	Reads procar and store
	useful informations
	"""
    def __init__(self, filename: str):
        """
			Args:
				filename (str): name of the procar file in VASP
			Members:
				num_kpoints: Number of kpoints used in the simulation
				num_bands: Number of bands used in the simulation
				size_procar_header: Number of lines of the procar header in PROCAR file
		"""
        self.filename = filename
        self.num_kpoints, self.num_bands = self._get_num_kpoints_and_bands()
        self.size_procar_header = 3
        self.is_spin_polarized = self._is_spin_polarized()
        if self.is_spin_polarized:
            logger.error("The calculation is spin polarized")
            raise ValueError(
                "Minushalf CLI does not support calculations with spin polarized"
            )

    def _is_spin_polarized(self) -> bool:
        """
        Check if the calculation
        is spin polarized
        """
        header_regex = re.compile(
            r"^\s*#\s+of\s+k-points:\s+([0-9]+)\s+#\s+of\s+bands:\s+([0-9]+)")
        with open(self.filename, "r") as procar:

            for line in islice(procar, self.size_procar_header, None):
                header_match = header_regex.match(line)
                if header_match:
                    return True
        return False

    def _get_num_kpoints_and_bands(self) -> tuple:
        """
			Returns:
				(kpoints,band_number) (tuple): The number of kpoints
				used in the simulation and the number of bands used
				for each kpoint.
		"""
        regex = re.compile(
            r"^\s*#\s+of\s+k-points:\s+([0-9]+)\s+#\s+of\s+bands:\s+([0-9]+)")
        number_kpoints = None
        number_bands = None
        with open(self.filename) as procar:
            for line in procar:
                if regex.match(line):
                    number_kpoints = regex.match(line).group(1)
                    number_bands = regex.match(line).group(2)
                    return (int(number_kpoints), int(number_bands))

    def get_band_projection(self, kpoint: int, band_number: int):
        """
		Get the band projection for an specific kpoint and number of band
			Args:
				kpoint (int): Number of kpoints
				band_number (int): Number of the band
		"""

        projections = {}
        kpoint_regex = re.compile(r"^\s*k-point\s*([0-9]+)\s+")
        band_regex = re.compile(r"^\s*band\s*([0-9]+)")
        projections_regex = re.compile(
            r"^\s*([0-9]+)\s*([-+]?[0-9]*\.?[0-9]+)")

        with open(self.filename, "r") as procar:
            current_kpoint = None
            current_band = None

            for line in islice(procar, self.size_procar_header, None):

                kpoint_match = kpoint_regex.match(line)
                if kpoint_match:
                    current_kpoint = int(kpoint_match.group(1))

                band_match = band_regex.match(line)
                if band_match:
                    current_band = int(band_match.group(1))

                if current_kpoint == kpoint and current_band == band_number:
                    if projections_regex.match(line):
                        parsed_line = line.split()
                        projections["{}".format(parsed_line[0])] = [
                            float(elem) for elem in parsed_line[1:-1]
                        ]
                elif current_kpoint and current_kpoint > kpoint:
                    return projections

        return projections



"""
Default cli messages
"""
import click
from pyfiglet import Figlet


def welcome_message(text: str) -> None:
    """
    Print welcome message
    """
    renderizer = Figlet()
    click.echo(renderizer.renderText(text))


def end_message() -> None:
    """
    Print end message
    """
    renderizer = Figlet()
    click.echo(renderizer.renderText("END"))



"""
Enum type for correction codes used
in minushalf.yaml
"""
from enum import Enum, unique


@unique
class CorrectionCode(Enum):
    """
    Enum type for the correction codes
    """

    v = "v"       # Simple valence correction
    vf = "vf"     # Fractional valence correction
    c = "c"       # Simple conduction correction
    cf = "cf"     # Fractional conduction correction
    vc = "vc"     # Simple valence and simple conduction corrections
    vfc = "vfc"   # Valence fractional and simple conduction corrections
    vcf = "vcf"   # Simple valence  and conduction fractional corrections
    vfcf = "vfcf" # Valence fractionary and conduction fractional corrections

    def __str__(self):
        return str(self.name)

    @staticmethod
    def to_list():
        """
        Generate list of available correction codes
        """
        return list(map(lambda element: element.value, CorrectionCode))



"""
List the available methods to obtain cut initial guess from the
nearest neighbor distance.
"""
from enum import Enum, unique


@unique
class CutInitialGuessMethods(Enum):
    """
    Enum for different methods of obtaining the initial cut guess.
    """

    three_dimensions = "3d"  # three dimensional crystals

    def __str__(self):
        return str(self.name)

    @staticmethod
    def to_list():
        """
        Generate list of available methods
        """
        return list(map(lambda element: element.value, CutInitialGuessMethods))



"""
Parse valence orbital line in INP file.
"""


def parse_valence_orbitals(line: str) -> dict:
    """
    Parse valence orbital line in principal
    quantum number, angular momentum quantum number
    and eletronic occupation

        Args:
            line (str): line of imp file that represents
            a valence orbital
        Returns:
            A dictionary with fields n, l and eletronic occupation
        """
    parsed_line = line.split()
    orbital = {
        "n": int(parsed_line[0]),
        "l": int(parsed_line[1]),
        "occupation": [float(value) for value in parsed_line[2:]],
    }

    return orbital



"""
Trimming function
"""
import numpy as np
from minushalf.data import Constants


def trimming_function(
    radius: np.array,
    ion_potential: np.array,
    atom_potential: np.array,
    cut: float,
    amplitude: float,
) -> np.array:
    r"""
        Function that generate the potential for fractional occupation. The potential
        is cuted by a a function theta(r) to avoid divergence in calculations.
        The function of potential is defined as follows:

        .. math::
            V_{1/2} = (V_{atom}- V_{ion})\cdot \theta (r)

        where theta is:


        .. math::
            \theta (r)=A\cdot (1-(\frac{r}{CUT})^{n})^{3},r\leq CUT


        .. math::
            \theta (r) = 0,  r > CUT

        ---

            Args:
                cut (float): cutting parameter to cancel the potential

                amplitude (float): multiplicative factor of the potential function

                radius (np.array): rays in which the potential was calculated

                ion_potential (np.array): Atom pseudopotential with fractional occupation

                atom_potential (np.array): Atom pseudopotential with all electrons

            Returns:
                potential of fractional electron occupation at the exact level to be corrected


        """
    const = Constants()

    potential = np.array(
        4 * const.pi_constant * const.rydberg *
        np.power(const.bohr_radius, 3) *
        np.power(1 - np.power(radius / cut, const.trimming_exponent), 3) *
        (ion_potential - atom_potential) * amplitude)

    potential[radius >= cut] = 0

    return potential



"""
Leads with input file (INP.ae)
read by atomic program.
"""
import numpy as np
import fortranformat as ff
from minushalf.data import (ElectronicDistribution, PeriodicTable,
                            ExchangeCorreltion, CalculationCode)
from .drop_comments import drop_comments
from .parse_valence_orbital_line import parse_valence_orbitals


class InputFile:
    """
    Parses input file.
    """
    def __init__(self,
                 exchange_correlation_type: str,
                 calculation_code: str,
                 chemical_symbol: str,
                 esoteric_line: str,
                 number_valence_orbitals: int,
                 number_core_orbitals: int,
                 valence_orbitals: list,
                 description: str = "",
                 last_lines: list = None) -> None:
        """
        Args:
            chemical_symbol (str): Symbol of the chemical element (H, He, Li...)

            esoteric_line (str):  Its use is somewhat esoteric and for most
            calculations it should contain just a 0.0 in the position shown.

            exchange_correlation_type (str): functional of exchange and correlation
            ((r)ca(s), (r)wi(s), (r)hl(s), (r)gl(s) ,(r)bh(s), (r)pb(s), (r)rp(s), (r)rv(s), (r)bl(s))

            calculation code (str): Calculation code for inp file (ae)

            number_valence_orbitals (int): Number of orbitals in valence

            number_core_orbitals (int): Number of orbitals in the core

            valence_orbitals (list): list of dictionaries with the following
            properties: {"n": principal quantum number,"l":secondary quantum number,
            "occupation": occupation in the level}

            last_lines (list): any line or property that comes after
            electronic distribution
        """
        self.exchange_correlation_type = exchange_correlation_type
        self.calculation_code = calculation_code
        self.description = description
        self.chemical_symbol = chemical_symbol
        self.esoteric_line = esoteric_line
        self.number_core_orbitals = number_core_orbitals
        self.number_valence_orbitals = number_valence_orbitals
        self.valence_orbitals = valence_orbitals
        if not last_lines:
            self.last_lines = []
        else:
            self.last_lines = last_lines

    @property
    def chemical_symbol(self) -> str:
        """
        Returns:
            Chemical symbol of the element (H, He, Li...)
        """
        return self._chemical_symbol

    @chemical_symbol.setter
    def chemical_symbol(self, symbol: str) -> None:
        """
        Verify if the symbol is a valid periodic table element and
        format the string correctly.

        Args:
            symbol (str): chemical symbol of the element (H, He, Li...)
        """

        try:
            PeriodicTable[symbol]
        except KeyError as symbol_not_found:
            raise ValueError("The chemical symbol passed is not correct"
                             ) from symbol_not_found

        self._chemical_symbol = symbol.capitalize()

    @property
    def exchange_correlation_type(self) -> str:
        """
        Returns:
            Functional of exchange and correlation
            (ca, wi, hl, gl,bh, pb, rp, rv, bl
        """
        return self._exchange_correlation_type

    @exchange_correlation_type.setter
    def exchange_correlation_type(self,
                                  exchange_correlation_type: str) -> None:
        """
        Verify if the functional of  exchange and correlation is valid
        conforms the ATOM documentation

        Args:
            exchange_correlation_type (str): functional of exchange and correlation
            (ca, wi, hl, gl ,bh, pb, rp, rv, bl()
        """
        try:
            value = ExchangeCorreltion[exchange_correlation_type].value
        except KeyError as code_not_found:
            raise KeyError(
                "Your value of exchange and correlation functional is not valid"
            ) from code_not_found

        self._exchange_correlation_type = value

    @property
    def calculation_code(self) -> str:
        """
        Returns:
            Calculation code for inp file (ae)
        """
        return self._calculation_code

    @calculation_code.setter
    def calculation_code(self, calculation_code: str) -> None:
        """
        Verify if the calculation is valid
        conforms the ATOM documentation

        Args:
            calculation code (str): Calculation code for inp file (ae)
        """
        try:
            value = CalculationCode[calculation_code].value
        except KeyError as code_not_found:
            raise KeyError(
                "Your value of calculation is not valid") from code_not_found

        self._calculation_code = value

    def electron_occupation(self, electron_fraction: float,
                            secondary_quantum_number: int) -> None:
        """
        Corrects the input file of the atomic program,
        decreasing a fraction of the electron in a
        layer specified by the secondary quantum number

            Args:
                electron_fraction (float): Fraction of the electron
                that will be decreased in the INP file. Can vary between 0 and 0.5

                secondary_quantum_number (int): Specifies the layer on which
                the occupation is to be made.
        """
        for value in reversed(self.valence_orbitals):

            is_equal = np.isclose(value["occupation"][0],
                                  0.0,
                                  rtol=1e-04,
                                  atol=1e-08,
                                  equal_nan=False)

            if (not is_equal and value["l"] == secondary_quantum_number):
                value["occupation"][0] -= electron_fraction
                break
        else:
            raise Exception(
                "Trouble with occupation. Please verify the parameters passed and the INP file."
            )

    def to_stringlist(self) -> list:
        """
            Returns:
                List with the lines of the INP file.
        """
        input_lines = []

        input_lines.append("{:<3}{}{:<6}{}\n".format("", self.calculation_code,
                                                     "", self.description))

        chemical_symbol_line = "n={}".format(self.chemical_symbol)
        exchange_correlation_line = "c={}".format(
            self.exchange_correlation_type)

        second_line_formater = ff.FortranRecordWriter('1x,a4,1x,a4,2x')
        if len(self.chemical_symbol) == 1:
            second_line_formater = ff.FortranRecordWriter('1x,a3,2x,a4,2x')

        input_lines.append("{}\n".format(
            second_line_formater.write(
                [chemical_symbol_line, exchange_correlation_line])))

        input_lines.append(self.esoteric_line)

        orbital_numbers_formater = ff.FortranRecordWriter('2i5')
        orbital_numbers = orbital_numbers_formater.write(
            [self.number_core_orbitals, self.number_valence_orbitals])

        input_lines.append("{}\n".format(orbital_numbers))

        for orbital in self.valence_orbitals:
            quantum_number_formater = ff.FortranRecordWriter('2i5')
            quantum_numbers = quantum_number_formater.write(
                [orbital["n"], orbital["l"]])

            occupation_formater = ff.FortranRecordWriter('2f10.3')
            occupation = occupation_formater.write(orbital["occupation"])

            input_lines.append("{}{}\n".format(quantum_numbers, occupation))

        for line in self.last_lines:
            input_lines.append(line)

        return input_lines

    def to_file(self, filename: str = "./INP") -> None:
        """
        Write INP file
            Args:
                filename (str): name of the output file
        """

        with open(filename, "w") as input_file:
            lines = self.to_stringlist()
            input_file.writelines(lines)

    @staticmethod
    def from_file(filename: str = "./INP") -> any:
        """
        Parse INP.ae file.

            Args:
                filename: name of the INP file.
            Returns:
                input_file: instance of InputFile class.
        """
        with open(filename) as input_file:

            lines_without_comments = drop_comments(input_file.readlines())

            try:
                calculation_code = lines_without_comments[0].split()[0]
                description = " ".join(lines_without_comments[0].split()[1:])
            except ValueError as bad_inp_format:
                raise ValueError("Description or calculation code not provided"
                                 ) from bad_inp_format

            try:
                chemical_symbol = lines_without_comments[1].split()[0].split(
                    "=")[1]
                exchange_correlation_type = lines_without_comments[1].split(
                )[1].split("=")[1]
            except ValueError as bad_inp_format:
                raise ValueError(
                    "Chemical symbol or exchange correlation not provided"
                ) from bad_inp_format

            esoteric_line = lines_without_comments[2]

            try:
                number_core_orbitals = int(
                    lines_without_comments[3].split()[0])
                number_valence_orbitals = int(
                    lines_without_comments[3].split()[1])
            except ValueError as bad_inp_format:
                raise ValueError(
                    "Number of core orbitals or number of valence orbitals not provided"
                ) from bad_inp_format

            try:
                valence_orbitals = [
                    parse_valence_orbitals(lines_without_comments[i])
                    for i in range(4, 4 + number_valence_orbitals)
                ]
            except ValueError as bad_inp_format:
                raise ValueError("Valence orbitals do not provided correctly"
                                 ) from bad_inp_format

            last_lines = lines_without_comments[4 + number_valence_orbitals:]

            return InputFile(exchange_correlation_type, calculation_code,
                             chemical_symbol, esoteric_line,
                             number_valence_orbitals, number_core_orbitals,
                             valence_orbitals, description, last_lines)

    @staticmethod
    def minimum_setup(chemical_symbol: str,
                      exchange_correlation_type: str,
                      maximum_iterations: int = 100,
                      calculation_code: str = "ae") -> any:
        """
        Create INP file with minimum setup.

            Args:
            chemical_symbol (str): Symbol of the chemical element (H, He, Li...).

            exchange_correlation_type (str): functional of exchange and correlation
            ( ca, wi, hl, gl, bh, pb, rp , rv, bl)

            maximum_iterations (int): Maximum number of iterations for atomic program.
            The default is 100

            Returns:
                input_file: instance of InputFile class.
        """
        description = "{}".format(chemical_symbol)
        esoteric_line = "{:<7}0.0{:<7}0.0{:<7}0.0{:<7}0.0{:<7}0.0{:<7}0.0\n".format(
            '', '', '', '', '', '')
        last_lines = ["{} maxit\n".format(maximum_iterations)]

        try:
            electronic_distribution = ElectronicDistribution[
                chemical_symbol].value
        except ValueError as element_not_found:
            raise ValueError("This element its not available in our database"
                             ) from element_not_found
        number_core_orbitals = int(electronic_distribution[0].split()[0])
        number_valence_orbitals = int(electronic_distribution[0].split()[1])
        valence_orbitals = [
            parse_valence_orbitals(orbital)
            for orbital in electronic_distribution[1:]
        ]

        return InputFile(exchange_correlation_type, calculation_code,
                         chemical_symbol, esoteric_line,
                         number_valence_orbitals, number_core_orbitals,
                         valence_orbitals, description, last_lines)



"""
Correct crystal potential to
fractional occupations
"""
import copy
import numpy as np
from minushalf.interfaces import PotentialFile
from .vtotal import Vtotal
from .trimming_function import trimming_function
from .correct_potential_fourier_transform import correct_potential_fourier_transform


class AtomicPotential():
    """
    Correct atomic potential fourier tranform
    for fractional occupations in valence
    or conduction bands
    """
    def __init__(
        self,
        vtotal: Vtotal,
        vtotal_occupied: Vtotal,
        potential_file: PotentialFile,
    ) -> None:
        """
            Args:
                vtotal (Vtotal): Class of the atom's VTOTAL file
                with all electrons in the ground state

                vtotal_occupied (Vtotal):Class of the atom's VTOTAL file
                with fractional occupation

                input_file (InputFile): Class of the INP file used
                to generate the VTOTAL with fractional occupation

                potcar (Potcar): Class of the POTCAR file of the
                respective atom

        """
        self.vtotal = vtotal
        self.vtotal_occupied = vtotal_occupied
        self.potential_file = potential_file

    def occupy_potential(self, cut: float, amplitude) -> list:
        """
            Args:
                cut (float): Cutting parameter to cancel the potential

                amplitude (float): Multiplicative factor of the potential function

            Returns:
                A list that contains the potentials of fractional electron
                occupation at the exact level to be corrected.
        """
        trimmed_half_electron_potential = trimming_function(
            np.array(self.vtotal.radius, dtype=float),
            np.array(self.vtotal_occupied.down_potential, dtype=float),
            np.array(self.vtotal.down_potential, dtype=float),
            cut,
            amplitude,
        )
        return trimmed_half_electron_potential

    def correct_potential(self,
                          cut: float,
                          amplitude: float,
                          is_conduction: bool = False) -> list:
        """
        Correct fourier transform of the potential (V(k)) present in POTCAR file.

            Args:
                cut (float): Cutting parameter to cancel the potential

                amplitude (float): Multiplicative factor of the potential function

                is_conduction (bool): Indicates whether the potential correction will be
                made in the valence or in the conduction

            Returns:
                List of corrected potentials fourier transform


        """
        if is_conduction:
            amplitude = abs(amplitude) * -1

        occupied_potential = self.occupy_potential(cut, amplitude)

        ## generate a sample of the absolute values of the wave vector
        size_potential_sampling = len(
            self.potential_file.get_potential_fourier_transform())
        maximum_absolute_value_wave_vector = self.potential_file.get_maximum_module_wave_vector(
        )
        absolute_wave_vectors = np.arange(size_potential_sampling) * (
            maximum_absolute_value_wave_vector / size_potential_sampling)

        corrected_potential = correct_potential_fourier_transform(
            coefficient=np.array(
                self.potential_file.get_potential_fourier_transform()),
            k=absolute_wave_vectors,
            rays=np.array(self.vtotal.radius, dtype=float),
            occupation_potential=np.array(occupied_potential, dtype=float),
            cut=cut,
        )

        return corrected_potential

    def correct_file(
        self,
        potential: list,
        cut: float,
        amplitude: float,
        is_conduction: bool = False,
    ) -> None:
        """
        Create the potential file corrected

            Args:
                potential (list): List of corrected potentials fourier transform

                cut (float): Cutting parameter to cancel the potential

                amplitude (float): Multiplicative factor of the potential function
        """
        if is_conduction:
            amplitude = abs(amplitude) * -1

        filename = ""
        if np.isclose(amplitude, 1.0):
            filename = "{}cut{:.2f}".format(self.potential_file.get_name(),
                                            cut)
        else:
            filename = "{}cut{:.2f}A{:.1f}".format(
                self.potential_file.get_name(), cut, amplitude)

        copy_potcar = copy.deepcopy(self.potential_file)
        copy_potcar.potential = potential
        copy_potcar.to_file(filename)

    def get_corrected_file_lines(
        self,
        potential: list,
    ) -> list:
        """
        Create the potential file corrected

            Args:
                potential (list): List of corrected potentials fourier transform
            Returns:
                potential_lines(list): A List of potcar lines
        """
        copy_potcar = copy.deepcopy(self.potential_file)
        copy_potcar.potential = potential
        return copy_potcar.to_stringlist()



"""
Reads Outcar file, an output of
VASP software
"""
import re
import math
from collections import defaultdict


class Outcar():
    """
    Reads output informations stored
    in OUTCAR file
    """
    def __init__(self, filename: str):
        """
            Args:
                filename (str): name of the OUTCAR file in VASP

            Members:
                relative_distances (defaultdict(list)): An dictionary where
                the keys are the ion index given by VASP and the values are lists of tuples
                containing the index other ion and the relative distance
                respectively.
        """
        self.filename = filename
        self.relative_distances = self._get_distances()

    def nearest_neighbor_distance(self, ion_index: str) -> float:
        """
        Given the ion index, it returns the distance of the nearest neighbor
        to this ion

            Args:
                ion_index (str): The index of the ion given by VASP.

            Returns:
                nearest_neighbor_distance (float): The distance of the nearest neighbor.
        """
        distances = self.relative_distances[ion_index]
        nearest_distance = math.inf
        for distance in distances:
            nearest_distance = min(distance[1], nearest_distance)
        return nearest_distance

    def number_of_equal_neighbors(self, atoms_map: dict, symbol: str) -> int:
        """
        Given an map that links atoms symbols with it's index
        this function returns the number of neighbors of the atom with
        equal symbol but different indexes.

            Args:
                atoms_map (dict): Map the atoms index to their symbol.
                symbom (str): The symbol of the target atom.

            Returns:
                number_equal_neighbors (int): Returns the number of neighbors with
                                        same symbol but different indexes.
        """
        ion_index = None
        for key, value in atoms_map.items():
            if value == symbol:
                ion_index = key
                break

        number_equal_neighbors = 0
        ## Already listed neighbors
        visited_neighbors = {key: False for key in atoms_map.keys()}
        for element in self.relative_distances[ion_index]:
            if element[0] != ion_index and atoms_map[str(
                    element[0])] == symbol and not visited_neighbors[str(
                        element[0])]:
                visited_neighbors[str(element[0])] = True
                number_equal_neighbors += 1

        return number_equal_neighbors

    def _get_distances(self) -> defaultdict(list):
        """
            Returns:
                relative_distances (defaultdict(list)): An dictionary where
                the keys are the ion index given by VASP and the values are lists of tuples
                containing the index other ion and the relative distance
                respectively.
        """
        relative_distances = defaultdict(list)
        start_capture_regex = re.compile(
            r"\s*ion\s+position\s+nearest\s+neighbor\s+table")
        distances_line_regex = re.compile(
            r"\s+([0-9]).*(?=-)-\s+([0-9]\s+[0-9]*\.[0-9]+\s*)+")
        with open(self.filename) as outcar:
            start_capture = False
            for line in outcar:

                if start_capture and distances_line_regex.match(line):
                    ion_index = distances_line_regex.match(line).group(1)

                    ion_relative_distances_line = line.split("-")[1]
                    ion_relative_distances_and_index = re.findall(
                        r"[0-9]\s+[0-9]*\.[0-9]+\s*",
                        ion_relative_distances_line)

                    for element in ion_relative_distances_and_index:
                        element_cleaned = element.rstrip()
                        index = int(element_cleaned.split()[0])
                        distance = float(element_cleaned.split()[1])
                        relative_distances[ion_index].append((index, distance))
                elif start_capture:
                    break

                if start_capture_regex.match(line):
                    start_capture = True

        return relative_distances



"""
Transform informations about band_project generated
by cbm_character, vbm_character or band_character in
a normalized datafame Grouped by orbital types
"""
from collections import defaultdict
import pandas as pd
from minushalf.data import Orbital, OrbitalType


def projection_to_df(projection: defaultdict(list)) -> pd.DataFrame:
    """
    Transform received dictionaries into information with a higher degree of readability.

        Args:
            projection (defaultdict(list)): A dictionary containing the projections
            of each atom per orbital
        Returns:
            prolection_df (pd.DataFrame): A dataframe containing the projections
            per orbital type and normalized between 0 - 100
    """
    orbitals = [orbital.__str__() for orbital in Orbital]
    number_columns = 0
    for value in list(projection.values()):
        number_columns = max(number_columns, len(value))

    if not number_columns:
        raise ValueError('The projection vector is empty')

    if number_columns > len(orbitals):
        raise ValueError(
            'the projection vector has more columns than necessary')

    projection_df = (pd.DataFrame.from_dict(projection,
                                            orient="index",
                                            columns=orbitals[:number_columns]))
    normalized_df = projection_df.div(projection_df.sum().sum())
    normalized_df = normalized_df.mul(100)

    orbital_type = [str(elem) for elem in OrbitalType]

    def join_orbitals(cols):
        return [
            cat for col in cols for cat in orbital_type if col.startswith(cat)
        ]

    normalized_df = normalized_df.groupby(join_orbitals(projection_df.columns),
                                          axis=1).sum().round()
    return normalized_df



"""
Correction interface
"""
from abc import ABC, abstractmethod


class Correction(ABC):
    """
    Interface to correction
    algorithms
    """
    @abstractmethod
    def execute(self):
        """
        Execute the correction
        """



"""
Enum class for all elements
of the periodic table
"""
from enum import Enum, unique


@unique
class PeriodicTable(Enum):
    """Enum type for the elements of the periodic table."""
    H = "H"
    He = "He"
    Li = "Li"
    Be = "Be"
    B = "B"
    C = "C"
    N = "N"
    O = "O"
    F = "F"
    Ne = "Ne"
    Na = "Na"
    Mg = "Mg"
    Al = "Al"
    Si = "Si"
    P = "P"
    S = "S"
    Cl = "Cl"
    Ar = "Ar"
    K = "K"
    Ca = "Ca"
    Sc = "Sc"
    Ti = "Ti"
    V = "V"
    Cr = "Cr"
    Mn = "Mn"
    Fe = "Fe"
    Co = "Co"
    Ni = "Ni"
    Cu = "Cu"
    Zn = "Zn"
    Ga = "Ga"
    Ge = "Ge"
    As = "As"
    Se = "Se"
    Br = "Br"
    Kr = "Kr"
    Rb = "Rb"
    Sr = "Sr"
    Y = "Y"
    Zr = "Zr"
    Nb = "Nb"
    Mo = "Mo"
    Tc = "Tc"
    Ru = "Ru"
    Rh = "Rh"
    Pd = "Pd"
    Ag = "Ag"
    Cd = "Cd"
    In = "In"
    Sn = "Sn"
    Sb = "Sb"
    Te = "Te"
    I = "I"
    Xe = "Xe"
    Cs = "Cs"
    Ba = "Ba"
    La = "La"
    Ce = "Ce"
    Pr = "Pr"
    Nd = "Nd"
    Pm = "Pm"
    Sm = "Sm"
    Eu = "Eu"
    Gd = "Gd"
    Tb = "Tb"
    Dy = "Dy"
    Ho = "Ho"
    Er = "Er"
    Tm = "Tm"
    Yb = "Yb"
    Lu = "Lu"
    Hf = "Hf"
    Ta = "Ta"
    W = "W"
    Re = "Re"
    Os = "Os"
    Ir = "Ir"
    Pt = "Pt"
    Au = "Au"
    Hg = "Hg"
    Tl = "Tl"
    Pb = "Pb"
    Bi = "Bi"
    Po = "Po"
    At = "At"
    Rn = "Rn"
    Fr = "Fr"
    Ra = "Ra"
    Ac = "Ac"
    Th = "Th"
    Pa = "Pa"
    U = "U"
    Np = "Np"
    Pu = "Pu"
    Am = "Am"
    Cm = "Cm"
    Bk = "Bk"
    Cf = "Cf"
    Es = "Es"
    Fm = "Fm"
    Md = "Md"
    No = "No"
    Lr = "Lr"
    Rf = "Rf"
    Db = "Db"
    Sg = "Sg"
    Bh = "Bh"
    Hs = "Hs"
    Mt = "Mt"
    Ds = "Ds"
    Rg = "Rg"
    Cn = "Cn"
    Nh = "Nh"
    Fl = "Fl"
    Mc = "Mc"
    Lv = "Lv"
    Ts = "Ts"
    Og = "Og"



"""
Interface for
classes that handles with fourier
transforms of the potential
"""
from abc import ABC, abstractmethod


class PotentialFile(ABC):
    """
    Interface for defining common methods
    between classes that handles with potential
    files of different softwares
    """
    @abstractmethod
    def to_stringlist(self) -> list:
        """
        Abstract methods that returns a list
        containing the lines of the potential file
        """

    @abstractmethod
    def to_file(self, filename: str) -> None:
        """
        Abstract methods that output
        a potential file with a specific
        filename
        """

    @abstractmethod
    def get_potential_fourier_transform(self) -> list:
        """
        Abstract methods returns
        the fourier transform of the
        potential
        """

    @abstractmethod
    def get_maximum_module_wave_vector(self) -> float:
        """
        Abstract methods returns
        the maximum modulus of
        the wave vector in reciprocal space
        """

    @abstractmethod
    def get_name(self) -> float:
        """
        Abstract methods returns
        the name of the potential file
        """



"""
Implements the algorithm that automates the process
of vasp correction and optimizes the necessary parameters.
"""
import os
import shutil
from subprocess import Popen, PIPE
from pandas.core import base
from scipy.optimize import minimize
from loguru import logger
import pandas as pd
from minushalf.data import (CorrectionDefaultParams,
                            AtomicProgramDefaultParams, OrbitalType,
                            CutInitialGuessMethods)
from minushalf.utils import (InputFile, Vtotal, MinushalfYaml, AtomicPotential,
                             BandStructure, find_negative_band_gap,
                             CutInitialGuess)
from minushalf.interfaces import (Correction, Runner, SoftwaresAbstractFactory)


class VaspCorrection(Correction):
    """
    An algorithm that realizes corrections for
    VASP software
    """
    def __init__(
        self,
        root_folder: str,
        software_factory: SoftwaresAbstractFactory,
        runner: Runner,
        minushalf_yaml: MinushalfYaml,
        band_projection: pd.DataFrame,
        atoms: list,
        is_conduction: bool,
        correction_indexes: dict,
        only_conduction: bool = False,
    ):
        """
        init method for the vasp correction class
            Args:
                root_folder (str): Path to the folder where the  correction will be made for each atom

                atoms (list): Atoms name

                potential_filename (str): Name of the potential file used by each
                                          software that performs ab initio calculations

                band_projection (pd.DataFrame): Shows the contribution of each atom in the CBM or VBM

                runner (Runner): class to execute the program that makes ab initio calculations

                only_conduction (bool): Conduction correction without previous valence correction

                inplace (bool): Realize calculations inplace, do not create find_cut folders
        """
        self.root_folder = root_folder

        self.atoms = atoms

        self.potential_filename = software_factory.get_potential_class(
        ).get_name()

        self.band_projection = band_projection

        if is_conduction and not only_conduction:
            self.potential_folder = "corrected_valence_potfiles"
        else:
            self.potential_folder = minushalf_yaml.correction[
                CorrectionDefaultParams.potfiles_folder.name]

        self.exchange_correlation_type = minushalf_yaml.atomic_program[
            AtomicProgramDefaultParams.exchange_correlation_code.name]

        self.max_iterations = minushalf_yaml.atomic_program[
            AtomicProgramDefaultParams.max_iterations.name]

        self.calculation_code = minushalf_yaml.atomic_program[
            AtomicProgramDefaultParams.calculation_code.name]

        self.amplitude = minushalf_yaml.correction[
            CorrectionDefaultParams.amplitude.name]

        if is_conduction:
            self.cut_initial_guess = minushalf_yaml.correction[
                CorrectionDefaultParams.conduction_cut_guess.name]
        else:
            self.cut_initial_guess = minushalf_yaml.correction[
                CorrectionDefaultParams.valence_cut_guess.name]

        self.automatic_cut_guess = False
        if not self.cut_initial_guess:
            self.automatic_cut_guess = True

        self.cut_guesser = CutInitialGuess()

        self.tolerance = minushalf_yaml.correction[
            CorrectionDefaultParams.tolerance.name]

        self.runner = runner

        self.software_factory = software_factory

        self.atom_potential = None

        self.sum_correction_percentual = 100

        if is_conduction:
            self.corrected_potfiles_folder = "corrected_conduction_potfiles"
        else:
            self.corrected_potfiles_folder = "corrected_valence_potfiles"

        if is_conduction:
            self.correction_type = "conduction"
        else:
            self.correction_type = "valence"

        self.is_conduction = is_conduction

        self.correction_indexes = correction_indexes

        self.input_files = ["INCAR", "POSCAR", "KPOINTS"]

        self.inplace = minushalf_yaml.correction[
            CorrectionDefaultParams.inplace.name]

    @property
    def potential_folder(self) -> str:
        """
        Returns:
            Name of the folder that helds all the potential files
            initially not corrected.
        """
        return self._potential_folder

    @potential_folder.setter
    def potential_folder(self, path: str) -> None:
        """
        Verify if the folder exists and contains all files needed

        Args:
            path (str): Path of the folder that helds all the potential files
            initially not corrected.
        """

        for atom in self.atoms:
            filename = "{}.{}".format(self.potential_filename.upper(),
                                      atom.lower())
            abs_path = os.path.join(path, filename)
            if not os.path.exists(abs_path):
                logger.error("Potential folder incomplete")
                raise FileNotFoundError("Potential folder lacks of files.")

        self._potential_folder = path

    def execute(self) -> tuple:
        """
        Execute vasp correction algorithm
        """
        ## make (valence|conduction) folder in mkpotfiles
        self.root_folder = os.path.join(self.root_folder, self.correction_type)
        if os.path.exists(self.root_folder):
            shutil.rmtree(self.root_folder)
        os.mkdir(self.root_folder)

        cuts_per_atom_orbital = {}
        self._make_corrected_potential_folder()

        self.sum_correction_percentual = self._get_sum_correction_percentual()

        for symbol, orbitals in self.correction_indexes.items():
            cut = self._find_best_correction(symbol, orbitals)
            cuts_per_atom_orbital[symbol] = cut

        gap = self._get_result_gap()
        return (cuts_per_atom_orbital, gap)

    def _get_result_gap(self) -> float:
        """
        Return the gap after the optimization of all potfiles
        """
        calculation_folder = "calculate_{}_gap".format(self.correction_type)
        if os.path.exists(calculation_folder):
            shutil.rmtree(calculation_folder)
        os.mkdir(calculation_folder)

        for file in self.input_files:
            shutil.copyfile(file, os.path.join(calculation_folder, file))

        potfile_path = os.path.join(calculation_folder,
                                    self.potential_filename)
        potential_file = open(potfile_path, "w")
        try:
            for atom in self.atoms:
                atom_potfilename = "{}.{}".format(
                    self.potential_filename.upper(), atom.lower())
                atom_potpath = os.path.join(self.corrected_potfiles_folder,
                                            atom_potfilename)
                with open(atom_potpath) as file:
                    potential_file.write(file.read())
        finally:
            potential_file.close()
        ## Run ab initio calculations
        self.runner.run(calculation_folder)

        eigenvalues = self.software_factory.get_eigenvalues(
            base_path=calculation_folder)
        fermi_energy = self.software_factory.get_fermi_energy(
            base_path=calculation_folder)
        atoms_map = self.software_factory.get_atoms_map(
            base_path=calculation_folder)
        num_bands = self.software_factory.get_number_of_bands(
            base_path=calculation_folder)
        band_projection_file = self.software_factory.get_band_projection_class(
            base_path=calculation_folder)

        band_structure = BandStructure(eigenvalues, fermi_energy, atoms_map,
                                       num_bands, band_projection_file)

        gap_report = band_structure.band_gap()
        return gap_report["gap"]

    def _make_corrected_potential_folder(self):
        """
        create the folder that will store
        the potfiles corrected with an cut
        value that returns tha maximum gap
        """
        name = self.corrected_potfiles_folder
        if os.path.exists(name):
            shutil.rmtree(name)
        os.mkdir(name)
        for atom in self.atoms:
            try:
                potential_filename = "{}.{}".format(
                    self.potential_filename.upper(), atom.lower())
                potential_path = os.path.join(self.potential_folder,
                                              potential_filename)
                potential_file = open(potential_path, "r")

                new_potential_path = os.path.join(name, potential_filename)
                new_potential_file = open(new_potential_path, "w")
                new_potential_file.write(potential_file.read())

            finally:
                potential_file.close()
                new_potential_file.close()

    def _get_sum_correction_percentual(self) -> float:
        """
        Sum of the percentuals of orbitals
        that will be corrected.
        """
        total_sum = 0
        for symbol, orbitals in self.correction_indexes.items():
            for orbital in orbitals:
                total_sum += self.band_projection[orbital][symbol]

        if total_sum == 0:
            logger.error(
                "No orbital selected for correction. Check you treshhold")
            raise ValueError(
                "No orbital selected for correction. Check you treshhold")

        return total_sum

    def _find_best_correction(self, symbol: str, orbitals: list) -> float:
        """
        Correct the potcar of the atom symbol in the
        orbital given. Then, find the best cut to the
        the correciton.

            Args:
                symbol (str): Atom symbol
                orbital (str): Orbital type (s,p,d,f)
            Returns:
                gap_and_cut (tuple): Tuple containing
                the optimum cut and the gap generated
                by the correction.
        """
        folder_name = "mkpotcar_{}".format(symbol.lower())
        path = os.path.join(self.root_folder, folder_name)
        if os.path.exists(path):
            shutil.rmtree(path)
        os.mkdir(path)
        self._generate_atom_potential(path, symbol)

        percentuals = {}
        ## Check for bonds with equal atoms
        atoms_map = self.software_factory.get_atoms_map()
        number_equal_neighbors = self.software_factory.get_number_of_equal_neighbors(
            atoms_map=atoms_map, symbol=symbol)
        for orbital in orbitals:
            value = (100 / (1 + number_equal_neighbors)) * (
                self.band_projection[orbital][symbol] /
                self.sum_correction_percentual)
            percentuals[orbital] = round(value)
        self._generate_occupation_potential(path, percentuals)

        self.atom_potential = self._get_atom_potential_class(path, symbol)

        cut = self._find_cut(symbol, path)
        self._write_result_in_potfile(symbol, cut, self.amplitude)
        return cut

    def _write_result_in_potfile(
        self,
        symbol: str,
        cut: float,
        amplitude: float,
    ):
        """
        Write the result of the optimization in the
        corrected potfiles folder.

            Args:
                symbol (str): Atom symbol
                amplitude (float): Scale factor to trimming function
                cut (float): Cut radius to the algorithm

        """
        potential = self.atom_potential.correct_potential(
            cut, amplitude, is_conduction=self.is_conduction)
        lines = self.atom_potential.get_corrected_file_lines(potential)

        new_potential_path = os.path.join(
            self.corrected_potfiles_folder,
            "{}.{}".format(self.potential_filename.upper(), symbol.lower()))

        if os.path.exists(new_potential_path):
            os.remove(new_potential_path)

        with open(new_potential_path, "w") as file:
            file.writelines(lines)

    def _generate_atom_potential(
        self,
        base_path: str,
        symbol: str,
    ) -> None:
        """
        Make a dir with the atoms name,generate
        the input file and run the atomic program

            Args:
                symbol (str): Atom symbol
                base_path (str): Path to mkpotcar{symbol}
        """

        folder_path = os.path.join(base_path, "pseudopotential")
        if os.path.exists(folder_path):
            shutil.rmtree(folder_path)
        os.mkdir(folder_path)
        input_file = InputFile.minimum_setup(
            symbol,
            self.exchange_correlation_type,
            self.max_iterations,
            self.calculation_code,
        )
        input_file.to_file(os.path.join(folder_path, "INP"))
        process = Popen(['minushalf', 'run-atomic', "--quiet"],
                        stdout=PIPE,
                        stderr=PIPE,
                        cwd=folder_path)

        _, stderr = process.communicate()
        if stderr:

            raise Exception("Call to atomic program failed")

    def _generate_occupation_potential(
        self,
        base_path: str,
        percentuals: dict,
    ) -> None:
        """
        Generate the potential for the occupation
        of a fraction of half electron.

            Args:
                base_path (str): Path to mkpotcar{symbol}
                percentual (dict): Dict where the key is the orbital type
                and the value is the percentual to be corrected
        """
        folder_path = os.path.join(base_path, "pseudopotential")
        if not os.path.exists(folder_path):

            raise FileNotFoundError(
                "Folder for pseudopotential does not exist")
        secondary_quantum_numbers = ",".join(
            [str(OrbitalType[key].value) for key in percentuals.keys()])

        joined_percentual = ",".join(
            [str(value) for value in percentuals.values()])

        process = Popen([
            "minushalf", "occupation", "{}".format(secondary_quantum_numbers),
            "{}".format(joined_percentual), "--quiet"
        ],
                        stdout=PIPE,
                        stderr=PIPE,
                        cwd=folder_path)

        _, stderr = process.communicate()
        if stderr:
            raise Exception("Call to occupation command failed")

    def _get_atom_potential_class(self, base_path: str,
                                  symbol: str) -> AtomicPotential:
        """
        Creates atom_potential class

            Args:
                symbol (str): Atom symbol
                base_path (str): Path to mkpotcar{symbol}
        """
        pseudopotential_folder = os.path.join(base_path, "pseudopotential")
        vtotal_path = os.path.join(pseudopotential_folder, "VTOTAL.ae")
        vtotal_occ_path = os.path.join(pseudopotential_folder, "VTOTAL_OCC")
        vtotal = Vtotal.from_file(vtotal_path)
        vtotal_occ = Vtotal.from_file(vtotal_occ_path)

        potential_filename = "{}.{}".format(self.potential_filename.upper(),
                                            symbol.lower())

        potential_class = self.software_factory.get_potential_class(
            potential_filename, self.corrected_potfiles_folder)

        return AtomicPotential(vtotal, vtotal_occ, potential_class)

    def _find_cut(self, symbol: str, base_path: str) -> tuple:
        """
        Find the cut which gives the maximum gap

            Args:
                symbol (str): Atom symbol
                base_path (str): Path to mkpotcar{symbol}
        """
        if not self.inplace:
            folder = os.path.join(base_path, "find_cut")
            if os.path.exists(folder):
                shutil.rmtree(folder)
            os.mkdir(folder)
        else:
            base_path = '.'  # Local Path

        function_args = {
            "base_path": base_path,
            "software_factory": self.software_factory,
            "runner": self.runner,
            "symbol": symbol,
            "default_potential_filename": self.potential_filename,
            "atom_potential": self.atom_potential,
            "potfiles_folder": self.potential_folder,
            "amplitude": self.amplitude,
            "atoms": self.atoms,
            "software_files": self.input_files,
            "is_conduction": self.is_conduction,
            "inplace": self.inplace
        }
        if self.automatic_cut_guess:
            atoms_map = self.software_factory.get_atoms_map()
            ion_index = None
            for key, value in atoms_map.items():
                if value == symbol:
                    ion_index = key
                    break

            nearest_distance = self.software_factory.get_nearest_neighbor_distance(
                ion_index)
            self.cut_initial_guess = self.cut_guesser.guess(
                nearest_distance,
                CutInitialGuessMethods.three_dimensions.value)

        result = minimize(find_negative_band_gap,
                          x0=self.cut_initial_guess,
                          args=(function_args),
                          method="Nelder-Mead",
                          options={'xatol': self.tolerance})

        if not result.success:
            logger.error("Optimization failed")
            raise Exception("Optimization failed.")

        cut = result.x[0]
        return cut



"""
List atomic orbitals and
their respective groups
"""
from enum import Enum, unique


@unique
class OrbitalType(Enum):
    """
    Enum type for orbital type. Indices are basically the azimuthal quantum
    number, l.
    """

    s = 0
    p = 1
    d = 2
    f = 3

    def __str__(self):
        return str(self.name)


@unique
class Orbital(Enum):
    """
    Enum type for orbitals. The indices are basically the order in
    which the orbitals are reported in VASP and has no special meaning.
    """

    s = 0
    py = 1
    pz = 2
    px = 3
    dxy = 4
    dyz = 5
    dz2 = 6
    dxz = 7
    dx2 = 8
    f_3 = 9
    f_2 = 10
    f_1 = 11
    f0 = 12
    f1 = 13
    f2 = 14
    f3 = 15

    def __str__(self):
        return str(self.name)



"""
Runner abstract class
"""
from abc import ABC, abstractmethod


class Runner(ABC):
    """
    Run softwares that
    realizes ab initio calculations
    """
    @abstractmethod
    def run(self, cwd: str):
        """
        Create command and
        run the subprocess for it
        """



"""
Lists minushalf.yml parameters and their default values
"""
from enum import Enum, unique
from minushalf import commands
import aenum


@unique
class MinushalfParams(Enum):
    """
    minushalf.yaml parameters.
    """

    software = "software"
    atomic_program = "atomic_program"
    correction = "correction"

    def __str__(self):
        return str(self.name)

    @staticmethod
    def to_list():
        """
        Returns a list with the name of the parameters present in minushalf.yaml.
        """
        return list(map(lambda element: element.value, MinushalfParams))

    @staticmethod
    def to_dict():
        """
        Returns a dictionary with the name of the parameters present in minushalf.yaml.
        """
        values = map(lambda element: element.value, MinushalfParams)
        keys = map(lambda element: element.__str__(), MinushalfParams)
        return dict(zip(keys, values))


class CorrectionDefaultParams(aenum.Enum, settings=aenum.NoAlias):
    """
    Default value of parameters in the correction tag.
    """

    correction_code = "v"
    potfiles_folder = "minushalf_potfiles"
    amplitude = 1.0
    valence_cut_guess = None
    conduction_cut_guess = None
    tolerance = 0.01
    fractional_valence_treshold = 10
    fractional_conduction_treshold = 9
    overwrite_vbm = []
    overwrite_cbm = []
    inplace = False

    def __str__(self):
        return str(self.name)

    @staticmethod
    def to_list():
        """
        Returns a list of default parameters.
        """
        return list(map(lambda element: element.value,
                        CorrectionDefaultParams))

    @staticmethod
    def to_dict():
        """
        Returns a dictionary of default parameters.
        """
        values = map(lambda element: element.value, CorrectionDefaultParams)
        keys = map(lambda element: element.__str__(), CorrectionDefaultParams)
        return dict(zip(keys, values))


@unique
class VaspDefaultParams(Enum):
    """
    Default value of parameters in the vasp tag.
    """

    command = ['mpirun','vasp']

    def __str__(self):
        return str(self.name)

    @staticmethod
    def to_list():
        """
        Returns a list of default parameters.
        """
        return list(map(lambda element: element.value, VaspDefaultParams))

    @staticmethod
    def to_dict():
        """
        Returns a dictionary of default parameters.
        """
        values = map(lambda element: element.value, VaspDefaultParams)
        keys = map(lambda element: element.__str__(), VaspDefaultParams)
        return dict(zip(keys, values))


@unique
class AtomicProgramDefaultParams(Enum):
    """
    Default value of parameters in the atomic_program tag.
    """

    exchange_correlation_code = "pb"
    calculation_code = "ae"
    max_iterations = 100

    def __str__(self):
        return str(self.name)

    @staticmethod
    def to_list():
        """
        Returns a list of default parameters.
        """
        return list(
            map(lambda element: element.value, AtomicProgramDefaultParams))

    @staticmethod
    def to_dict():
        """
        Returns a dictionary of default parameters.
        """
        values = map(lambda element: element.value, AtomicProgramDefaultParams)
        keys = map(lambda element: element.__str__(),
                   AtomicProgramDefaultParams)
        return dict(zip(keys, values))



"""
Software abstract Factory
"""
from abc import ABC, abstractmethod
from typing import List
from .potential_file import PotentialFile
from .band_projection_file import BandProjectionFile


class SoftwaresAbstractFactory(ABC):
    """
    Abstract Factory for create instances for
    each supported software.
    """
    @abstractmethod
    def get_atoms_map(self, filename: str, base_path: str = None) -> dict:
        """
        Abstract method for returns a map
        of the atomic symbol to its index.
        """

    @abstractmethod
    def get_band_projection_class(
        self,
        filename: str,
        base_path: str = None,
    ) -> BandProjectionFile:
        """
        Abstract method for returns the class that
        handles with the projections of atoms orbitals
        in the bands.
        """

    @abstractmethod
    def get_fermi_energy(self, filename: str, base_path: str = None) -> float:
        """
        Abstract method for returna
        energy of the fermi level.
        """

    @abstractmethod
    def get_number_of_bands(self, filename: str, base_path: str = None) -> int:
        """
        Abstract method for returns the number of bands used in the calculation
        """

    @abstractmethod
    def get_number_of_kpoints(self,
                              filename: str,
                              base_path: str = None) -> int:
        """
        Abstract method for returns the number of kpoints used in the calculation
        """

    @abstractmethod
    def get_potential_class(self,
                            filename: str,
                            base_path: str = None) -> PotentialFile:
        """
        Abstract method for returns the potential class
        """

    @abstractmethod
    def get_eigenvalues(self, filename: str, base_path: str = None) -> dict:
        """
        Abstract method for returns eigenvalues
        for each band and each kpoint
        """

    @abstractmethod
    def get_runner(self, command: List[str]):
        """
        Get a class that run
        the software for ab initio calculations
        """

    @abstractmethod
    def get_nearest_neighbor_distance(self,
                                      ion_index: str,
                                      filename: str,
                                      base_path: str = None) -> float:
        """
        Abstract method for returns the nearest neighbor distance for
        an ion in the solid.
        """

    @abstractmethod
    def get_number_of_equal_neighbors(self,
                                      atoms_map: dict,
                                      symbol: str,
                                      filename: str = "OUTCAR",
                                      base_path: str = None) -> float:
        """
        Given an map that links atoms symbols with it's index
        this function returns the number of neighbors of the atom with
        equal symbol but different indexes.

            Args:
                atoms_map (dict): Map the atoms index to their symbol.
                symbom (str): The symbol of the target atom.

            Returns:
                number_equal_neighbors (int): Returns the number of neighbors with
                                        same symbol but different indexes.
        """



"""
Init file for data module
"""
from .orbital import Orbital, OrbitalType
from .electronic_distribution import ElectronicDistribution
from .periodic_table import PeriodicTable
from .constants import Constants
from .softwares import Softwares
from .exchange_correlation import ExchangeCorreltion
from .calculation_code import CalculationCode
from .minushalf_yaml_default_configuration import (
    VaspDefaultParams,
    AtomicProgramDefaultParams,
    CorrectionDefaultParams,
    MinushalfParams,
)
from .correction_code import CorrectionCode
from .cut_initial_guess_methods import CutInitialGuessMethods



"""
Parse cut string used in correct potential file
"""
import numpy as np


def parse_cut(cut: str) -> list:
    """
    Parse cut in a list of numbers.

    Args:
        cut (str): Cut energy to be used in the program, it can be
                   passed in two ways:

                    unique value : float or integer
                    range:  begin(float|integer):pass(float|integer):end(float|integer)
    Returns:

        cut_numbers (list): Permited values of cut.
    """
    if not cut:
        raise ValueError('A valeu of cut energy must be provided.')

    cuts = cut.split(':')

    try:
        cut_number = [float(element) for element in cuts]
    except ValueError as bad_input:
        raise ValueError("Invalid Input.") from bad_input

    if len(cut_number) == 1:
        return cut_number
    if len(cut_number) != 3:
        raise ValueError()

    return np.arange(cut_number[0],
                     cut_number[2],
                     cut_number[1],
                     dtype=np.float64)



"""
List softwares supported
by the CLI
"""
from enum import Enum, unique


@unique
class Softwares(Enum):
    """
    Enum type for the softwares supported by the program
    """

    vasp = "VASP"

    def __str__(self):
        return str(self.name)

    @staticmethod
    def to_list():
        """
        Generate list of available softwares
        """
        return list(map(lambda element: element.value, Softwares))



"""
Init file for utils module
"""
from .input_file import InputFile
from .vtotal import Vtotal
from .cli_messages import welcome_message, end_message
from .drop_comments import drop_comments
from .parse_valence_orbital_line import parse_valence_orbitals
from .trimming_function import trimming_function
from .correct_potential_fourier_transform import correct_potential_fourier_transform
from .projection_to_df import projection_to_df
from .band_structure import BandStructure
from .atomic_potential import AtomicPotential
from .check_file_exists import (check_eigenval_exists, check_vasprun_exists,
                                check_procar_exists, check_potcar_exists,
                                check_outcar_exists)
from .parse_cut import parse_cut
from .minushalf_yaml import MinushalfYaml
from .make_minushalf_results import make_minushalf_results
from .negative_band_gap import find_negative_band_gap
from .fractionary_correction_indexes import get_fractionary_correction_indexes
from .simple_correction_indexes import get_simple_correction_indexes
from .cut_initial_guess import CutInitialGuess



"""
Reads and analyze POTCAR file
"""
import re
from itertools import chain
import fortranformat as ff
import numpy as np
from minushalf.interfaces import PotentialFile


class Potcar(PotentialFile):
    """
    Parse the POTCAR file, a vasp input file. It store
    the fourier coefficients and the restant lines of the file
    """
    def __init__(self, filename: str = "POTCAR") -> None:
        self.filename = filename
        self.psctr_parameters = self._get_psctr_parameters()
        self.k_max_text, self.potential = self._get_potential()
        self.k_max = float(self.k_max_text)
        self.last_lines = self._get_last_lines()
        self.name = "POTCAR"

    def get_potential_fourier_transform(self) -> list:
        """
        Returns:
            potential(list): List of fourier transform
            of the potential
        """
        return self.potential

    def get_name(self) -> str:
        """
        Returns potential file name
        """
        return self.name

    def get_maximum_module_wave_vector(self) -> float:
        """
        Returns:
            k_max (float):maximum modulus of the wave vector in reciprocal space
        """
        return self.k_max

    def to_stringlist(self) -> list:
        """
            Returns:
                potcar_lines (list): List of the POTCAR lines
        """
        fourier_coefficients_lines = [
            "{:<3}{}{:<5}\n".format('', self.k_max_text, '')
        ]
        try:
            grouped_coefficients = self.potential.reshape((-1, 5))
        except ValueError as bad_shape:
            raise ValueError(
                "Fourier coefficents not provided correctly") from bad_shape

        fortran_formater = ff.FortranRecordWriter('(1E26.8)')
        for group in grouped_coefficients:
            formated_numbers = []
            for number in group:
                formated_numbers.append(
                    fortran_formater.write([number]).strip())

            line = "{:<2}{}\n".format('', "  ".join(formated_numbers))
            fourier_coefficients_lines.append(line)

        lines = [
            self.psctr_parameters, fourier_coefficients_lines, self.last_lines
        ]
        return list(chain.from_iterable(lines))

    def to_file(self, filename: str) -> None:
        """
        Write POTCAR file
            Args:
                filename (str): Name of the file
        """
        lines = self.to_stringlist()
        with open(filename, "w") as new_potcar:
            new_potcar.writelines(lines)

    def _get_psctr_parameters(self) -> list:
        """
        Read PSCTR parameters in POTCAR

            Returns:
                psctr_parameters (list): A list with the lines that contain
                the psctr_patameters.
        """
        psctr_parameters = []
        regex_end_psctr_catch = re.compile(r"^\s*local\s+part")
        with open(self.filename, "r") as potcar:
            for line in potcar:
                psctr_parameters.append(line)
                if regex_end_psctr_catch.match(line):
                    return psctr_parameters

    def _get_potential(self) -> tuple:
        """
        Read the maximum fourier coefficient and the list
        of fourier coefficients present in POTCAR

            Returns:
                (k_max,potential) (tuple): A tuple
                containing the maximum modulus of the wave vector in
                reciprocal space string and a list with all fourier
                transforms for the potential.
        """
        k_max_text = ""
        potential = []
        regex_begin_catch = re.compile(r"^\s*local\s+part")
        regex_end_catch = re.compile(r"^\s*[a-z]+")

        with open(self.filename, "r") as potcar:
            ## Jump lines
            for line in potcar:
                if regex_begin_catch.match(line):
                    break

            k_max_text = potcar.readline().strip()

            for line in potcar:
                if regex_end_catch.match(line):
                    potential = list(chain.from_iterable(potential))
                    return (k_max_text, np.array(potential, dtype=np.float64))
                potential.append(line.split())

    def _get_last_lines(self) -> list:
        """
        Read the last lines in POTCAR

            Returns:
                last_lines (list): A list with the lines that are after
                the fourier coefficients. These lines will not be used
                for correction purposes
        """
        last_lines = []
        regex_begin_potential = re.compile(r"^\s*local\s+part")
        regex_end_potential = re.compile(r"^\s*[a-z]+")

        with open(self.filename, "r") as potcar:
            ## Jump potential lines
            for line in potcar:
                if regex_begin_potential.match(line):
                    break

            for line in potcar:
                if regex_end_potential.match(line):
                    last_lines.append(line)
                    break
            ## Catch all lines from here
            for line in potcar:
                last_lines.append(line)

        return last_lines



"""
Definition of the minushalf CLI
"""
import click
from minushalf.commands import (
    run_atomic,
    occupation,
    vbm_character,
    cbm_character,
    band_character,
    create_input,
    correct_potfile,
    band_gap,
    execute,
)


@click.group()
def minushalf():
    """
    CLI for automating DFT -1/2 calculations
    """


minushalf.add_command(run_atomic)
minushalf.add_command(occupation)
minushalf.add_command(vbm_character)
minushalf.add_command(cbm_character)
minushalf.add_command(band_character)
minushalf.add_command(band_gap)
minushalf.add_command(create_input)
minushalf.add_command(correct_potfile)
minushalf.add_command(execute)



"""
Makes fractional occupation on INP file
"""
import sys
import click
from loguru import logger
from minushalf.data import (CalculationCode, ExchangeCorreltion)
from minushalf.utils import (welcome_message, end_message, InputFile)


@click.command()
@click.argument('chemical_symbol', type=str, nargs=1)
@click.option('-e',
              '--exchange_correlation_code',
              type=click.Choice(ExchangeCorreltion.to_list(),
                                case_sensitive=False),
              nargs=1,
              default=ExchangeCorreltion.pb.value,
              show_default=True,
              help="""
            Represents the functional of exchange and correlation,it can assume the following values:


              ca: Ceperley-Alder

              wi: Wigner

              hl: Hedin-Lundqvist

              gl: Gunnarson-Lundqvist

              bh: Von Barth-Hedin

              pb: PBE scheme by Perdew, Burke, and Ernzerhof

              rp: RPBE scheme by Hammer, Hansen, and Norskov

              rv: revPBE scheme by Zhang and Yang

              bl: BLYP (Becke-Lee-Yang-Parr) scheme

              """)
@click.option('-c',
              '--calculation_code',
              type=click.Choice(CalculationCode.to_list(),
                                case_sensitive=False),
              nargs=1,
              default=CalculationCode.ae.value,
              show_default=True,
              help="""Represents calculation code,it can
              assume the following values:

              ae: All electrons
              """)
@click.option(
    '-m',
    "--maximum_iterations",
    type=click.IntRange(0, 10000),
    nargs=1,
    default=100,
    show_default=True,
    help="""Maximum number of iterations performed by the atomic program""")
@click.option('-f',
              "--filename",
              type=str,
              nargs=1,
              default='INP',
              show_default=True,
              help="""Name of the created file""")
@click.option('--quiet', default=False, is_flag=True)
def create_input(chemical_symbol: str, exchange_correlation_code: str,
                 calculation_code: str, maximum_iterations: int, filename: str,
                 quiet: bool):
    """
    Create the input file for the run-atomic command.


    Requires:

        CHEMICAL_SYMBOL: Chemical symbol of the atom (H, He, Na, Li...). Check the list
                         of available atoms in the docs


    Returns:

        INP: The input file for run-atomic command
    """

    welcome_message("minushalf")

    if quiet:
        logger.remove()
        logger.add(sys.stdout, level="ERROR")

    input_file = InputFile.minimum_setup(chemical_symbol.capitalize(),
                                         exchange_correlation_code,
                                         maximum_iterations, calculation_code)
    logger.info("Creating INP file")

    input_file.to_file(filename)

    end_message()



"""
Returns band-gap with the sinal changed, so
one can use minimization algorithms to find the
cut value that results in the maximum band_gap
"""
import os
import shutil
import math
from loguru import logger
from minushalf.interfaces import SoftwaresAbstractFactory
from .atomic_potential import AtomicPotential
from .band_structure import BandStructure


def _get_corrected_potfile_lines(
    cut: float,
    atom_potential: AtomicPotential,
    amplitude: float,
    is_conduction: bool,
) -> list:
    """
        Modify the potential file that was corrected

            Args:

                cut (float): Distance to trimm the potential
                atom_potential(AtomicPotential): Class that represents the
                                                 potential of the atom
                amplitude (float): Scale factor to trimming function
        """
    potential = atom_potential.correct_potential(cut,
                                                 amplitude,
                                                 is_conduction=is_conduction)
    lines = atom_potential.get_corrected_file_lines(potential)
    return lines


def _join_potfiles(
    base_path: str,
    default_potential_filename: str,
    atoms: list,
    potfiles_folder: str,
    symbol: str,
    corrected_potfile_lines: list,
) -> None:
    """
        Join valence potfiles in one

        Args:

            base_path (str): Path to mkpotcar{symbol}_{orbital}
            symbol (str): Atom symbol
            default_potential_filename (str): The default potential filename for each software
            potfiles_folder (str): Folder containing unmodified potfiles
            symbol (str): symbol of the atom to be corrected
            potfile_lines (list): Lines of the corrected potential file

        """
    path = os.path.join(base_path, default_potential_filename.upper())
    if os.path.exists(path):
        os.remove(path)
    potential_file = open(path, "w")

    try:
        for atom in atoms:

            if not atom == symbol:
                potential_filename = "{}.{}".format(
                    default_potential_filename.upper(), atom.lower())
                potential_path = os.path.join(potfiles_folder,
                                              potential_filename)

                with open(potential_path, "r") as file:
                    potential_file.write(file.read())
            else:
                potential_file.writelines(corrected_potfile_lines)
    finally:
        potential_file.close()


def _set_up_cut_folder(base_path: str, input_files: list, cut: float) -> str:
    """
        Creates and populates the folder where the first principles calculations will be done

        Args:
            base_path (str): Path where the folder will be created
            cut (float): Distance to trimm the potential
            input_files (List[str]): List of input files
        
        Returns:
            cut_folder (str): Path to folder where the first principles calculations will be done
    """
    cut_folder = _create_cut_folder(base_path, cut)
    _copy_input_files(input_files, cut_folder)
    return cut_folder


def _create_cut_folder(base_path: str, cut: float) -> str:
    """
        Creates the folder where the first principles calculations will be done

        Args:
            
            base_path (str): Path where the folder will be created
            cut (float): Distance to trimm the potential
        
        Returns:
            cut_folder (str): Path to folder where the first principles calculations will be done
    """
    find_cut_path = os.path.join(base_path, "find_cut")
    cut_folder = os.path.join(find_cut_path, "cut_{:.2f}".format(cut))

    if os.path.exists(cut_folder):
        shutil.rmtree(cut_folder)
    os.mkdir(cut_folder)

    return cut_folder


def _copy_input_files(input_files: list, destination_folder: str) -> None:
    """
        Copy the input files for the first principles calculations

        Args:
            input_files (List[str]): List of input files
            destination_folder (str): Path to which files will be copied
    """

    for file in input_files:
        shutil.copyfile(file, os.path.join(destination_folder, file))


def _get_gap(software_factory: SoftwaresAbstractFactory,
             cut_folder: str) -> float:
    """
        Returns the gap value

        Args:
            software_factory (SoftwaresAbstractFactory) : Get informations for output files of the first principle calculations
            cut_folder (str): Folder where first principles calculations were made
        
        Returns:
            gap (float): Gap of the semiconductor material
    """
    eigenvalues = software_factory.get_eigenvalues(base_path=cut_folder)
    fermi_energy = software_factory.get_fermi_energy(base_path=cut_folder)
    atoms_map = software_factory.get_atoms_map(base_path=cut_folder)
    num_bands = software_factory.get_number_of_bands(base_path=cut_folder)
    band_projection_file = software_factory.get_band_projection_class(
        base_path=cut_folder)

    band_structure = BandStructure(eigenvalues, fermi_energy, atoms_map,
                                   num_bands, band_projection_file)

    gap_report = band_structure.band_gap()
    return gap_report["gap"]


def find_negative_band_gap(cuts: list, *args: tuple) -> float:
    """
                Run vasp and return the gap value multiplied by -1

                Args:
                    cuts (float): List of cuts

                    args (tuple): tuple containning a dictionary with the fields
                                   base_path (str): Path to mkpotcar{symbol}_{orbital}
                                   symbol (str): Atom symbol
                                   default_potential_filename (str): The default potential filename for each software
                                   potfiles_folder (str): Folder containing unmodified potfiles
                                   amplitude (float): scale factor to trimming function
                                   runner (Runner): runner for the software
                                   software_factory(SoftwaresAbstractFactory): Factory for each software
                                   atom_potential(AtomicPotential): Holds fourier transforms of the potential
                                   software_files (list): Aditional files besides potential file to make ab initio calculations

                Returns:

                    negative_gap (float): band gap multiplied for -1
    """
    extra_args = args[0]
    cut = cuts[0]
    runner = extra_args["runner"]
    software_factory = extra_args["software_factory"]
    inplace = extra_args["inplace"]
    is_conduction = extra_args["is_conduction"]

    if cut <= extra_args["atom_potential"].vtotal.radius[1]:
        return math.inf

    ## Add condition to include inplace calculations
    if not inplace:
        cut_folder = _set_up_cut_folder(extra_args["base_path"],
                                        extra_args["software_files"], cut)
    else:
        cut_folder = '.'

    potfile_lines = _get_corrected_potfile_lines(
        cut,
        extra_args["atom_potential"],
        extra_args["amplitude"],
        extra_args["is_conduction"],
    )
    _join_potfiles(
        cut_folder,
        extra_args["default_potential_filename"],
        extra_args["atoms"],
        extra_args["potfiles_folder"],
        extra_args["symbol"],
        potfile_lines,
    )

    runner.run(cut_folder)

    gap = _get_gap(software_factory, cut_folder)

    ## Logger
    if is_conduction:
        logger.info(
            "CONDUCTION CORRECTION: Current CUT value is {:.2f} a.u".format(
                cut))
        logger.info(
            "CONDUCTION CORRECTION: Current Gap value is {:.2f} eV".format(
                gap))
    else:
        logger.info(
            "VALENCE CORRECTION: Current CUT value is {:.2f} a.u".format(cut))
        logger.info(
            "VALENCE CORRECTION: Current Gap value is {:.2f} eV".format(gap))

    if cut < 0.5:
        logger.warning(
            "Pay attention, CUT values less than 0.5 a.u might have no physical meaning."
        )

    return (-1) * gap



"""
Interfaces for files thar
contain band projections
"""
from abc import ABC, abstractmethod


class BandProjectionFile(ABC):
    """
    Class to handle files
    containing informations
    about projections of
    the atoms in bands.
    """
    @abstractmethod
    def get_band_projection(self, kpoint: int, band_number: int) -> dict:
        """
        Abstract method for return the
        contribution of each atom orbital in a
        specific band of an specific kpoint.
        """



"""
Gives cut initial guess
"""
from minushalf.data import CutInitialGuessMethods


def _3d_crystals(distance: float) -> float:
    """
        Gives cut initial guess for 3D crystals

            Args:
                distance (float): Nearest neighbor distance.

            Returns:
                cut_guess (float): An initial guess to cut.
        """
    return 0.15 + 0.84 * 1.88973 * distance


class CutInitialGuess:
    """
    Estimate cut inital guess from
    the nearest neighbor distance.
    """
    def __init__(self):
        """
        Describe options to guess
        """
        self.options = {
            CutInitialGuessMethods.three_dimensions.value: _3d_crystals,
        }

    def guess(self, distance: float, method: str) -> float:
        """
        Given the nearest neighbor distance and the method, it returns the
        initial guess.

            Args:
                method (str): method of gessing
                distance (float): nearest neighbor distance

            Returns:
                cut_guess (float): An initial guess to cut.
        """
        return self.options[method](distance)



"""
Aims to show how the last valence band are composed by the orbitals of each atom.
"""
import click
from minushalf.softwares import VaspFactory
from minushalf.data import Softwares
from minushalf.utils import (welcome_message, end_message, projection_to_df,
                             BandStructure)


@click.command()
@click.option(
    '-s',
    '--software',
    type=click.Choice(Softwares.to_list(), case_sensitive=False),
    default=Softwares.vasp.value,
    show_default=True,
    help="""Specifies the software used to perform ab initio calculations.""")
@click.option('-b',
              '--base-path',
              type=click.Path(),
              nargs=1,
              help="""Path to folder where the relevant files are located.""")
def vbm_character(software: str, base_path: str) -> None:
    """Uses output files from softwares that perform ab initio calculations to discover the
       last valence band (VBM) and extract, in percentage, its character corresponding to each
       orbital type (s, p, d, ... ). The names of the files required for each
       software are listed below, it is worth mentioning that their names cannot be modified.

    VASP: PROCAR, EIGENVAL, vasprun.xml
    """

    welcome_message("minushalf")

    softwares = {"VASP": VaspFactory()}

    factory = softwares[software.upper()]

    eigenvalues = factory.get_eigenvalues(base_path=base_path)
    fermi_energy = factory.get_fermi_energy(base_path=base_path)
    atoms_map = factory.get_atoms_map(base_path=base_path)
    num_bands = factory.get_number_of_bands(base_path=base_path)
    band_projection_file = factory.get_band_projection_class(
        base_path=base_path)

    band_structure = BandStructure(eigenvalues, fermi_energy, atoms_map,
                                   num_bands, band_projection_file)
    vbm_projection = band_structure.vbm_projection()
    normalized_df = projection_to_df(vbm_projection)

    click.echo(normalized_df.to_markdown())

    end_message()



"""
Parser for minushalf.yaml
"""
import yaml
from loguru import logger
from minushalf.data import (
    Softwares,
    VaspDefaultParams,
    AtomicProgramDefaultParams,
    CorrectionDefaultParams,
    CorrectionCode,
    MinushalfParams,
)


class MinushalfYaml():
    """
    Class that parses the input
    for the execute command
    """
    def __init__(
        self,
        software: str,
        software_configurations: dict,
        atomic_program: dict,
        correction: dict,
    ):
        """
        Constructs a class for input in the execute command
        """
        self.software = software.upper()
        self.software_configurations = software_configurations
        self.atomic_program = atomic_program
        self.correction = correction

    @property
    def software(self) -> str:
        """
        Returns:
            Name of the software used for ab initio calculations (VASP,...)
        """
        return self._software

    @software.setter
    def software(self, name: str) -> None:
        """
        Verify if the symbol is a valid periodic table element and
        format the string correctly.

        Args:
            symbol (str): chemical symbol of the element (H, He, Li...)
        """

        available_softwares = Softwares.to_list()
        is_software_avalilable = any(element == name
                                     for element in available_softwares)
        if not is_software_avalilable:
            logger.error("Parameter software is not filled correctly")
            raise ValueError("Parameter software is not filled correctly")

        self._software = name.upper()

    @property
    def software_configurations(self) -> str:
        """
        Returns:
            Some configurations for the softwares that runs
            ab initio calculations
        """
        return self._software_configurations

    @software_configurations.setter
    def software_configurations(self, configurations: dict) -> None:
        """
            Verify if the file has all the cofigurations needed. If not, use
            the default values
                Args: configurations (dict): Configuration parameters for each software
                    VASP:
                        number_of_cores: Number of cores to run the proccess. The
                        default is one.
                        path_to_vasp: Path to the VASP software.The deafault is 'vasp'
        """
        choice_params = {Softwares.vasp.value: VaspDefaultParams}

        software_params = choice_params[self.software]
        default_params = software_params.to_dict()

        if not configurations:
            configurations = default_params
        else:
            for name, value in default_params.items():
                if not name in configurations:
                    configurations[name] = value
            for name, value in configurations.items():
                if not name in default_params:
                    raise ValueError("param {} is not defined".format(name))

        self._software_configurations = configurations

    @property
    def atomic_program(self) -> dict:
        """
        Returns:
            Some configurations for the atomic program
        """
        return self._atomic_program

    @atomic_program.setter
    def atomic_program(self, configurations: dict) -> None:
        """
        Verify if the file has all the cofigurations needed. If not, use
        the default values

            Args:
                configurations(dict): Configurations params for
                atomc program. Contain the following fields:

                    exchange_correlation_type: Default is 'pb'

                    calculation_code: Default is 'ae'

                    max_iterations: Default is 100
        """
        default_params = AtomicProgramDefaultParams.to_dict()
        if not configurations:
            configurations = default_params
        else:
            for name, value in default_params.items():
                if not name in configurations:
                    configurations[name] = value
            for name, value in configurations.items():
                if not name in default_params:
                    raise ValueError("param {} is not defined".format(name))

        self._atomic_program = configurations

    @property
    def correction(self) -> dict:
        """
        Returns:
            Some configurations for the correction in the
            potential files
        """
        return self._correction

    @correction.setter
    def correction(self, configurations: dict) -> None:
        """
        Verify if the file has all the cofigurations needed. If not, use
        the default values

            Args:
                configurations(dict): Configurations params for
                correct potential. Contain the following fields:

                    correction_code: Default is 'v'

        """
        default_params = CorrectionDefaultParams.to_dict()
        if not configurations:
            configurations = default_params
        else:
            for name, value in default_params.items():
                if not name in configurations:
                    configurations[name] = value
            for name, value in configurations.items():
                if not name in default_params:
                    raise ValueError("param {} is not defined".format(name))

        available_correction_codes = CorrectionCode.to_list()
        is_code_available = any(element == configurations[
            CorrectionDefaultParams.correction_code.__str__()]
                                for element in available_correction_codes)
        if not is_code_available:
            raise ValueError("Invalid value for correction code")

        self._correction = configurations

    @staticmethod
    def from_file(filename: str = "minushalf.yaml"):
        """
        Receives a file and catch all the parameters
        presents in the documentation
        """
        parsed_input = {}
        try:
            with open(filename, "r") as file:
                parsed_input = yaml.load(file, Loader=yaml.FullLoader)
        except FileNotFoundError:
            logger.info("File not found, default parameters will be used")

        if MinushalfParams.software.value in parsed_input:
            software = parsed_input[MinushalfParams.software.value]
        else:
            software = Softwares.vasp.value

        if software.lower() in parsed_input:
            software_configurations = parsed_input[software.lower()]
        else:
            software_configurations = None

        if MinushalfParams.atomic_program.value in parsed_input:
            atomic_program = parsed_input[MinushalfParams.atomic_program.value]
        else:
            atomic_program = None

        if MinushalfParams.correction.value in parsed_input:
            correction = parsed_input[MinushalfParams.correction.value]
        else:
            correction = None

        return MinushalfYaml(
            software,
            software_configurations,
            atomic_program,
            correction,
        )



"""
Function to check if the line
a commentarie or not
"""
import re


def drop_comments(lines: list) -> list:
    """
    Function to remove comments from lines in a file

        Args:
            lines(list): list of file lines
        Returns:
            lines_without_comments (list): lines of the file without comments
    """

    comment_pattern = r"#.*"
    lines_without_comments = [
        re.sub(comment_pattern, "", line) for line in lines
    ]

    return list(
        filter(lambda x: x != "\n" and x.strip() != "",
               lines_without_comments))



"""
Get atoms to be corrected for
simple valence correction and simple
condunction correction
"""
import itertools
import math
import pandas as pd
import numpy as np


def get_simple_correction_indexes(band_projection: pd.DataFrame) -> dict:
    """
        Get dataframe index of the orbital which contributes more
        to (VBM|CBM)

            Returns:

                correction_indexes (dict):A dict wherw the keys are the atoms
                                          symbols and the value is a list with the orbitals type to be
                                          corrected.
                                          Ex:
                                          {
                                          'Ga': ['p','s'],
                                          'N' : ['d','f'],
                                          }

        """

    columns_name = band_projection.columns.tolist()
    rows_name = band_projection.index.tolist()

    maximum_projection = -math.inf

    for row, column in itertools.product(rows_name, columns_name):
        projection = band_projection[column][row]
        maximum_projection = max(projection, maximum_projection)
        if np.isclose(projection, maximum_projection):
            max_row = row
            max_column = column

    correction_indexes = {max_row: [max_column]}
    return correction_indexes



"""
Function to check if a file exists
"""
import os


def check_vasprun_exists(func):
    """
    Function decrator to check if a file exists
    """
    def func_wrapper(self,
                     filename: str = "vasprun.xml",
                     base_path: str = None):
        path = filename
        if base_path:
            path = os.path.join(base_path, filename)
        if not os.path.exists(path):
            raise ValueError("File {} does not exist".format(filename))
        return func(self, filename, base_path)

    return func_wrapper


def check_procar_exists(func):
    """
    Function decrator to check if a file exists
    """
    def func_wrapper(self, filename: str = "PROCAR", base_path: str = None):
        path = filename
        if base_path:
            path = os.path.join(base_path, filename)
        if not os.path.exists(path):
            raise ValueError("File {} does not exist".format(filename))
        return func(self, filename, base_path)

    return func_wrapper


def check_eigenval_exists(func):
    """
    Function decrator to check if a file exists
    """
    def func_wrapper(self, filename: str = "EIGENVAL", base_path: str = None):
        path = filename
        if base_path:
            path = os.path.join(base_path, filename)
        if not os.path.exists(path):
            raise ValueError("File {} does not exist".format(filename))
        return func(self, filename, base_path)

    return func_wrapper


def check_potcar_exists(func):
    """
    Function decrator to check if a file exists
    """
    def func_wrapper(self, filename: str = "POTCAR", base_path: str = None):
        path = filename
        if base_path:
            path = os.path.join(base_path, filename)
        if not os.path.exists(path):
            raise ValueError("File {} does not exist".format(filename))
        return func(self, filename, base_path)

    return func_wrapper


def check_outcar_exists(func):
    """
    Function decrator to check if a file exists
    """
    def func_wrapper(self, *args, **kwargs):
        kwargs["filename"] = "OUTCAR"
        path = kwargs["filename"]
        if "base_path" in kwargs:
            path = os.path.join(kwargs["base_path"], kwargs["filename"])
        if not os.path.exists(path):
            raise ValueError("File {} does not exist".format(
                kwargs["filename"]))
        return func(self, *args, **kwargs)

    return func_wrapper



"""
Fourier Transform
"""
import numpy as np
from minushalf.data import Constants


def correct_potential_fourier_transform(
    coefficient: np.array,
    k: np.array,
    rays: np.array,
    occupation_potential: np.array,
    cut: float,
) -> np.array:
    r"""
    The pseudopotential is given in terms of the radial distance, and is only defined for r >= 0,
    as expected. Since it is only evaluated inside an integral from 0 to infinity, it does not
    matter what values it assumes for r < 0. A natural choice is to define the function to be
    zero for negative values, but a more convenient choice is to choose
    v(-r)=-v(r) and n(-r)=-n(r), since purely real and odd functions have purely imaginary Fourier transforms.
    Let v' and n' be the odd extensions of the potential and the number density, respectively.

    .. math::
        E_{v} = \int_{0}^{\infty}v(r)n(r)dr = \int_{0}^{\infty}v'(r)n'(r)dr =
        \frac{1}{2}\cdot\int_{-\infty}^{\infty}v'(r)n'(r)dr = -\frac{1}{2}\cdot\int_{-\infty}^{\infty}V(k)N(k)dk


    On the third equalitty, we used the fact that the product of two odd
    functions is even, and in the last step we have applied Parseval's theorem,
    considering that the Fourier transforms are purely imaginary. Even though
    the function may not pass through the origin, we can still make an odd extension, by making it discontinuous.

    The data stored on POTCAR corresponds to the Fourier transform of the odd extension of v. It can be approximated
    by the summation on the right, where the prefactors were ommited.

    .. math::
        V(k) = i\cdot \sqrt{\frac{2}{\pi}}\cdot\int_{0}^{\infty}v(r)sin(b\cdot k\cdot r)dr
        \Rightarrow V(k)\sim \sum^{N_{r}}_{i=1}
        \frac{(v[i]\cdot sin(b\cdot k\cdot r[i])+v[i-1]\cdot sin(b\cdot k\cdot r[i-1]))}{2\cdot (r[i]-r[i-1])}

    Computes the opposite of the imaginary part of the j-th fourier transform coefficient
    through numerical integrationIndex zero stands for the r=DeltaR, and the function is
    assumed to be zero at the origin. Thus, the first trapezium of the numerical integration is
    degenerated to a triangulum, and its area must be calculated as so.

        Args:
            coefficient (np.array): Fourier transform of the potential for the atom in its ground state

            k (np.array): The wave vector in reciprocal space

            rays(np.array): List of rays on which pseudopotential calculations were made

            occupation_potential (np.arraygit): Potential of fractional electron occupation at the exact level to be corrected

            cut(float): Cutting parameter to cancel the potential

        Returns:
            Fourier transform of the potential for the state with fractional occupation of the crystal
    """
    const = Constants()
    k = np.reshape(k, (len(k), 1))

    if not k[0][0]:
        k[0][0] = 10**(-12)

    try:
        filter_rays = rays[np.where(rays < cut)]
    except ValueError as cut_error:
        raise ValueError(
            "the cut is smaller than all the rays passed") from cut_error

    radius = filter_rays[1:].astype(float)
    lazy_radius = filter_rays[:-1].astype(float)
    potential = occupation_potential[1:len(filter_rays)].astype(float)
    lazy_potential = occupation_potential[:len(filter_rays) - 1].astype(float)

    partial_fourier_sum = (
        (radius - lazy_radius) *
        (potential * np.sin(const.bohr_radius * k * radius) +
         lazy_potential * np.sin(const.bohr_radius * k * lazy_radius)) / 2)

    ## Sum all rows and transpose

    initial_term = (occupation_potential[0] *
                    np.sin(const.bohr_radius * k * rays[0]) / 2 * rays[0])

    ## Reshape vectors to give the correct output format
    initial_term = np.reshape(initial_term, (len(initial_term)))
    k = np.reshape(k, (len(k)))

    return (coefficient + (initial_term + partial_fourier_sum.sum(axis=1)) /
            (const.bohr_radius * k))



"""
Execute command
"""
import os
import sys
import shutil
from collections import OrderedDict
from typing import List
import numpy as np
import pandas as pd
import click
from loguru import logger
from minushalf.utils import (
    MinushalfYaml,
    BandStructure,
    projection_to_df,
    welcome_message,
    end_message,
    make_minushalf_results,
    get_fractionary_correction_indexes,
    get_simple_correction_indexes,
)
from minushalf.softwares import (VaspFactory)
from minushalf.corrections import (VaspCorrection)
from minushalf.data import (Softwares, CorrectionCode, CorrectionDefaultParams)
from minushalf.interfaces import (SoftwaresAbstractFactory)


def get_vbm_projection(factory: SoftwaresAbstractFactory) -> pd.DataFrame:
    """
    Returns vbm projection
    """
    eigenvalues = factory.get_eigenvalues()
    fermi_energy = factory.get_fermi_energy()
    atoms_map = factory.get_atoms_map()
    num_bands = factory.get_number_of_bands()
    band_projection_file = factory.get_band_projection_class()

    band_structure = BandStructure(eigenvalues, fermi_energy, atoms_map,
                                   num_bands, band_projection_file)
    vbm_projection = band_structure.vbm_projection()
    normalized_df = projection_to_df(vbm_projection)
    return normalized_df


def get_cbm_projection(factory: SoftwaresAbstractFactory) -> pd.DataFrame:
    """
    Returns cbm projection
    """
    eigenvalues = factory.get_eigenvalues()
    fermi_energy = factory.get_fermi_energy()
    atoms_map = factory.get_atoms_map()
    num_bands = factory.get_number_of_bands()
    band_projection_file = factory.get_band_projection_class()

    band_structure = BandStructure(eigenvalues, fermi_energy, atoms_map,
                                   num_bands, band_projection_file)
    cbm_projection = band_structure.cbm_projection()
    normalized_df = projection_to_df(cbm_projection)
    return normalized_df


def get_atoms_list(factory: SoftwaresAbstractFactory) -> list:
    """
    Returns atoms_list
    """
    atoms_map = factory.get_atoms_map()
    atoms = [atoms_map[key] for key in sorted(atoms_map)]
    return list(OrderedDict.fromkeys(atoms))


def overwrite_band_projection(
        band_location: List[int],
        factory: SoftwaresAbstractFactory) -> pd.DataFrame:
    """
    Overwrite values in VBM or CBM band projection

        Args:

            band_location (List[int]): Kpoint and band number, respectively.

            factory (SoftwaresAbstractFactory): Software factory to create band structure class

        Returns:

            modified_band_projection (pd.Dataframe): Band projection data frame
                                                    with the values overwrited.

    """
    eigenvalues = factory.get_eigenvalues()
    fermi_energy = factory.get_fermi_energy()
    atoms_map = factory.get_atoms_map()
    num_bands = factory.get_number_of_bands()
    band_projection_file = factory.get_band_projection_class()

    band_structure = BandStructure(eigenvalues, fermi_energy, atoms_map,
                                   num_bands, band_projection_file)
    vbm_projection = band_structure.band_projection(*band_location)
    normalized_df = projection_to_df(vbm_projection)
    return normalized_df


@click.command()
@click.option('--quiet', default=False, is_flag=True)
def execute(quiet: bool):
    """
    Uses the Nelder-Mead method to find
    the optimal values for the CUT(S) and,
    finally, find the corrected Gap value.
    This command uses external software to
    perform ab initio calculations, so it must
    be installed in order to perform the command.
    Check the docs for an list of the softwares supported
    by the CLI.


        Requires:


            minushalf.yaml : Parameters file. Check the docs
                            for a more detailed description.

            ab_initio_files: Files needed to perform the ab initio calculations.
                            They must be in the same directory as the input
                            file minushalf.yaml

            potential_folder: Folder with the potential files for each atom in
                            the crystal. The files must be named in the following pattern
                            ${POTENTIAL_FILE_NAME}.${LOWERCASE_CHEMICAL_SYMBOL}

        Returns:

            minushalf_results.dat : File that contains the optimal
                                values of the cuts and the final
                                value of the Gap.

            corrected_valence_potfiles: Potential files corrected with opti-mum valence cuts.

            corrected_conduction_potfiles: Potential files corrected with optimum conduction cuts.
    """
    welcome_message("minushalf")

    if quiet:
        logger.remove()
        logger.add(sys.stdout, level="ERROR")
    ## Read yaml file
    logger.info("Reading minushalf.yaml file")
    minushalf_yaml = MinushalfYaml.from_file()
    correction_factory_chooser = {Softwares.vasp.value: VaspCorrection}
    software_factory_chooser = {Softwares.vasp.value: VaspFactory()}

    correction = correction_factory_chooser[minushalf_yaml.software]
    software_factory = software_factory_chooser[minushalf_yaml.software]

    ## Makes abinition calculation
    logger.info("Running ab initio calculations")
    software_configurations = minushalf_yaml.software_configurations
    runner = software_factory.get_runner(**software_configurations)
    runner.run()

    ## Makes root folder
    logger.info("Make potfiles folder")
    root_folder = "mkpotfiles"
    if os.path.exists(root_folder):
        shutil.rmtree(root_folder)
    os.mkdir(root_folder)

    ## get the bands projections (VBM and CBM)
    logger.info("Get VBM and CBM projections")
    vbm_projection = get_vbm_projection(software_factory)
    cbm_projection = get_cbm_projection(software_factory)

    ### Overwrite band projections
    if len(minushalf_yaml.correction[
            CorrectionDefaultParams.overwrite_vbm.name]) == 2:
        logger.warning(
            "You're changing directly the band character. This is not recommendend unless you know exactly what are you doing."
        )
        vbm_projection = overwrite_band_projection(
            minushalf_yaml.correction[
                CorrectionDefaultParams.overwrite_vbm.name], software_factory)
    if len(minushalf_yaml.correction[
            CorrectionDefaultParams.overwrite_cbm.name]) == 2:
        logger.warning(
            "You're changing directly the band character. This is not recommendend unless you know exactly what are you doing."
        )
        cbm_projection = overwrite_band_projection(
            minushalf_yaml.correction[
                CorrectionDefaultParams.overwrite_cbm.name], software_factory)

    ## get atoms list
    logger.info("Get atoms list")
    atoms = get_atoms_list(software_factory)

    ## amplitude logger
    if not np.isclose(
            minushalf_yaml.correction[CorrectionDefaultParams.amplitude.name],
            CorrectionDefaultParams.amplitude.value):
        logger.warning(
            "Amplitude value is different from 1.0. This is not recommended unless you know exactly what you are doing."
        )

    valence_options = {
        "root_folder": root_folder,
        "software_factory": software_factory,
        "runner": runner,
        "minushalf_yaml": minushalf_yaml,
        "band_projection": vbm_projection,
        "atoms": atoms,
        "is_conduction": False,
    }
    conduction_options = {
        "root_folder": root_folder,
        "software_factory": software_factory,
        "runner": runner,
        "minushalf_yaml": minushalf_yaml,
        "band_projection": cbm_projection,
        "atoms": atoms,
        "is_conduction": True
    }

    logger.info("Doing corrections")
    if minushalf_yaml.correction["correction_code"] == CorrectionCode.v.name:
        valence_options["correction_indexes"] = get_simple_correction_indexes(
            vbm_projection)
        valence_correction = correction(**valence_options)
        valence_cuts, valence_gap = valence_correction.execute()
        make_minushalf_results(valence_cuts=valence_cuts, gap=valence_gap)

    elif minushalf_yaml.correction[
            "correction_code"] == CorrectionCode.vf.name:

        valence_options[
            "correction_indexes"] = get_fractionary_correction_indexes(
                vbm_projection,
                treshold=minushalf_yaml.
                correction["fractional_valence_treshold"])

        valence_fractionary_correction = correction(**valence_options)
        valence_cuts, valence_gap = valence_fractionary_correction.execute()
        make_minushalf_results(valence_cuts=valence_cuts, gap=valence_gap)

    elif minushalf_yaml.correction[
            "correction_code"] == CorrectionCode.vc.name:
        valence_options["correction_indexes"] = get_simple_correction_indexes(
            vbm_projection)
        conduction_options[
            "correction_indexes"] = get_simple_correction_indexes(
                cbm_projection)
        valence_correction = correction(**valence_options)
        valence_cuts, _ = valence_correction.execute()
        conduction_correction = correction(**conduction_options)
        conduction_cuts, conduction_gap = conduction_correction.execute()
        make_minushalf_results(valence_cuts=valence_cuts,
                               gap=conduction_gap,
                               conduction_cuts=conduction_cuts)

    elif minushalf_yaml.correction[
            "correction_code"] == CorrectionCode.vfc.name:

        valence_options[
            "correction_indexes"] = get_fractionary_correction_indexes(
                vbm_projection,
                treshold=minushalf_yaml.
                correction["fractional_valence_treshold"])

        conduction_options[
            "correction_indexes"] = get_simple_correction_indexes(
                cbm_projection)

        valence_fractionary_correction = correction(**valence_options)
        valence_cuts, _ = valence_fractionary_correction.execute()
        conduction_correction = correction(**conduction_options)
        conduction_cuts, conduction_gap = conduction_correction.execute()
        make_minushalf_results(valence_cuts=valence_cuts,
                               gap=conduction_gap,
                               conduction_cuts=conduction_cuts)

    elif minushalf_yaml.correction[
            "correction_code"] == CorrectionCode.vfcf.name:

        valence_options[
            "correction_indexes"] = get_fractionary_correction_indexes(
                vbm_projection,
                treshold=minushalf_yaml.
                correction["fractional_valence_treshold"])

        conduction_options[
            "correction_indexes"] = get_fractionary_correction_indexes(
                cbm_projection,
                treshold=minushalf_yaml.
                correction["    "])

        valence_fractionary_correction = correction(**valence_options)
        valence_cuts, _ = valence_fractionary_correction.execute()
        conduction_fractionary_correction = correction(**conduction_options)
        conduction_cuts, conduction_gap = conduction_fractionary_correction.execute(
        )
        make_minushalf_results(valence_cuts=valence_cuts,
                               gap=conduction_gap,
                               conduction_cuts=conduction_cuts)

    elif minushalf_yaml.correction[
            "correction_code"] == CorrectionCode.vcf.name:
        valence_options["correction_indexes"] = get_simple_correction_indexes(
            vbm_projection)

        conduction_options[
            "correction_indexes"] = get_fractionary_correction_indexes(
                cbm_projection,
                treshold=minushalf_yaml.
                correction["fractional_conduction_treshold"])

        valence_correction = correction(**valence_options)
        valence_cuts, _ = valence_correction.execute()
        conduction_fractionary_correction = correction(**conduction_options)
        conduction_cuts, conduction_gap = conduction_fractionary_correction.execute(
        )
        make_minushalf_results(valence_cuts=valence_cuts,
                               gap=conduction_gap,
                               conduction_cuts=conduction_cuts)
    elif minushalf_yaml.correction["correction_code"] == CorrectionCode.c.name:
        conduction_options[
            "correction_indexes"] = get_simple_correction_indexes(
                cbm_projection)
        conduction_options["only_conduction"] = True

        conduction_simple_correction = correction(**conduction_options)
        conduction_cuts, conduction_gap = conduction_simple_correction.execute(
        )
        make_minushalf_results(gap=conduction_gap,
                               conduction_cuts=conduction_cuts)
    elif minushalf_yaml.correction[
            "correction_code"] == CorrectionCode.cf.name:
        conduction_options[
            "correction_indexes"] = get_fractionary_correction_indexes(
                cbm_projection,
                treshold=minushalf_yaml.
                correction["fractional_conduction_treshold"])
        conduction_options["only_conduction"] = True

        conduction_fractionary_correction = correction(**conduction_options)
        conduction_cuts, conduction_gap = conduction_fractionary_correction.execute(
        )
        make_minushalf_results(gap=conduction_gap,
                               conduction_cuts=conduction_cuts)
    end_message()



"""
Band structure informations
"""
from math import inf
from collections import defaultdict
import numpy as np
from minushalf.interfaces import BandProjectionFile


class BandStructure():
    """
    Extact band structure insights from
    VASP classes
    """
    def __init__(self, eigenvalues: dict, fermi_energy: float, atoms_map: dict,
                 num_bands: int, band_projection: BandProjectionFile):
        """
            Args:
                filename (str): name of the EIGENVAL file in VASP
            Members:
        """
        self.eigenvalues = eigenvalues
        self.fermi_energy = fermi_energy
        self.atoms_map = atoms_map
        self.num_bands = num_bands
        self.band_projection_file = band_projection

    def is_metal(self, tolerance: float = 1e-4) -> bool:
        """
        Check if the band structure indicates a metal by looking if the fermi
        level crosses a band.
        Returns:
            True if a metal, False if not
        """
        eigevalues = [*self.eigenvalues.values()]
        for index in range(self.num_bands):
            band_eigenvalues = np.array([array[index] for array in eigevalues],
                                        dtype=np.float64)
            if np.any(band_eigenvalues -
                      self.fermi_energy < -tolerance) and np.any(
                          band_eigenvalues - self.fermi_energy > tolerance):
                return True
        return False

    def vbm_index(self) -> tuple:
        """
        Find the kpoint and the band for vbm

            Returns:
                vbm_index (tuple): Contains the kpoint
                number and the band number of th vbm
        """
        if self.is_metal():
            raise Exception("Conduction band is not defined for metals")

        kpoint_vbm = None
        band_vbm = None
        max_energy_reached = -inf
        for kpoint, values in self.eigenvalues.items():
            for band_index, energy in enumerate(values):
                if max_energy_reached < energy < self.fermi_energy:
                    max_energy_reached = energy
                    kpoint_vbm = kpoint
                    band_vbm = band_index + 1

        return (kpoint_vbm, band_vbm)

    def cbm_index(self) -> tuple:
        """
        Find the kpoint and the band for cbm

            Returns:
                vbm_index (tuple): Contains the kpoint
                number and the band number of th vbm
        """
        if self.is_metal():
            raise Exception("Conduction band is not defined for metals")

        kpoint_cbm = None
        band_cbm = None
        min_energy_reached = inf
        for kpoint, values in self.eigenvalues.items():
            for band_index, energy in enumerate(values):
                if self.fermi_energy <= energy < min_energy_reached:
                    min_energy_reached = energy
                    kpoint_cbm = kpoint
                    band_cbm = band_index + 1
        return (kpoint_cbm, band_cbm)

    def vbm_projection(self) -> defaultdict(list):
        """
        Find the projection of each atom for valence
        band maximum.

            Returns:
                vbm_projection (defaultdict(list)): Contains the projection
                of each orbital of each atom in the respective band
        """
        vbm_index = self.vbm_index()
        procar_projection = self.band_projection_file.get_band_projection(
            *vbm_index)

        vbm_projection = {}
        for index, symbol in self.atoms_map.items():
            if vbm_projection.get(symbol, None):
                add_projections = np.add(procar_projection[index],
                                         vbm_projection[symbol])
                vbm_projection[symbol] = list(add_projections)
            else:
                vbm_projection[symbol] = procar_projection[index]
        return vbm_projection

    def cbm_projection(self) -> defaultdict(list):
        """
        Find the projection of each atom for valence
        band minimum.

            Returns:
                vbm_projection (defaultdict(list)): Contains the projection
                of each orbital of each atom in the respective band
        """
        cbm_index = self.cbm_index()

        procar_projection = self.band_projection_file.get_band_projection(
            *cbm_index)

        cbm_projection = {}
        for index, symbol in self.atoms_map.items():

            if cbm_projection.get(symbol, None):
                add_projections = np.add(procar_projection[index],
                                         cbm_projection[symbol])
                cbm_projection[symbol] = list(add_projections)
            else:
                cbm_projection[symbol] = procar_projection[index]

            cbm_projection[symbol] = procar_projection[index]
        return cbm_projection

    def band_projection(self, kpoint: int, band: int) -> defaultdict(list):
        """
        Find the projection of each atom for a specific band.
            Args:
                kpoint (int): Number of kpoints
                band_number (int): Number of the band

            Returns:
                band_projection (defaultdict(list)): Contains the projection
                of each orbital of each atom in the respective band
        """
        procar_projection = self.band_projection_file.get_band_projection(
            kpoint, band)

        band_projection = {}

        for index, symbol in self.atoms_map.items():
            if band_projection.get(symbol, None):
                add_projections = np.add(procar_projection[index],
                                         band_projection[symbol])
                band_projection[symbol] = list(add_projections)
            else:
                band_projection[symbol] = procar_projection[index]

        return band_projection

    def band_gap(self) -> dict:
        """
        Find VBM and CBM, then returns band gap
        Returns:
            VBM index and its eigenvalue, CBM index and its eigenvalue and band gap
        """
        vbm = self.vbm_index()
        vbm_eigenval = self.eigenvalues[vbm[0]][vbm[1] - 1]

        cbm = self.cbm_index()
        cbm_eigenval = self.eigenvalues[cbm[0]][cbm[1] - 1]

        gap_report = {
            "vbm":
            "VBM: Kpoint {}, band {} and eigenval {}".format(
                vbm[0], vbm[1], vbm_eigenval),
            "cbm":
            "CBM: Kpoint {}, band {} and eigenval {}".format(
                cbm[0], cbm[1], cbm_eigenval),
            "gap":
            cbm_eigenval - vbm_eigenval
        }
        return gap_report



"""
Implementation for vasp runner
"""
import subprocess
from typing import List
from minushalf.interfaces import Runner


class VaspRunner(Runner):
    """
    Output terminal command that
    aims to run vasp
    """
    def __init__(self, command: List[str]):
        """
        Args:
            path_to_vasp (str): Path to vasp binary
            number of cores (int): number of cores needed by the proccess
        """
        self.command = command

    def run(self, cwd: str = "."):
        """
        Create a subproccess to run
        vasp
        """
        subprocess.run(self.command, check=True, cwd=cwd)



"""
Command to read band gap
"""

import click
from minushalf.softwares import VaspFactory
from minushalf.data import Softwares
from minushalf.utils import (welcome_message, end_message, BandStructure)


@click.command()
@click.option(
    '-s',
    '--software',
    type=click.Choice(Softwares.to_list(), case_sensitive=False),
    default=Softwares.vasp.value,
    show_default=True,
    help="""Specifies the software used to perform ab initio calculations.""")
@click.option('-b',
              '--base-path',
              type=click.Path(),
              nargs=1,
              help="""Path to folder where the relevant files are located.""")
def band_gap(software: str, base_path: str) -> None:
    """Uses output files from softwares that perform ab initio calculations to
      provide the locations of VBM, CBM and the Gap value in electronvolts.The
      names of the files required for each software are listed below, it is
      worth mentioning that their names cannot be modified.

    VASP: PROCAR, EIGENVAL, vasprun.xml
    """

    welcome_message("minushalf")

    softwares = {"VASP": VaspFactory()}

    factory = softwares[software.upper()]

    eigenvalues = factory.get_eigenvalues(base_path=base_path)
    fermi_energy = factory.get_fermi_energy(base_path=base_path)
    atoms_map = factory.get_atoms_map(base_path=base_path)
    num_bands = factory.get_number_of_bands(base_path=base_path)
    band_projection_file = factory.get_band_projection_class(
        base_path=base_path)

    band_structure = BandStructure(eigenvalues, fermi_energy, atoms_map,
                                   num_bands, band_projection_file)

    gap_report = band_structure.band_gap()
    click.echo(gap_report["vbm"])
    click.echo(gap_report["cbm"])
    click.echo("Gap: {:.3f}eV".format(gap_report["gap"]))
    end_message()



"""
function make minushalf_results.dat
"""
from loguru import logger


def make_minushalf_results(
    gap: float,
    valence_cuts: dict = None,
    conduction_cuts: dict = None,
    name: str = "minushalf_results.dat",
) -> None:
    """
    Make output file for execute command, minushalf_results.dat.

        Args:
            gap (float): final gap in the correction method

            valence_cuts (dict):dictionary inform the atom symbol, orbital
                                and cut for valence correction in the following format:
                                {(symbol,orbital):cut}

            conduction_cuts (dict):dictionary inform the atom symbol, orbital
                                   and cut for condsuction correction in the following format.
                                   {(symbol,orbital):cut}

            name (str): name of the file
    """
    logger.info("Writing output file")
    with open(name, "w") as file:
        ## Write valence cuts
        if valence_cuts:
            file.write("Valence correction cuts:\n")
            for key, value in valence_cuts.items():
                symbol = key
                cut = value
                file.write("\t({}):{:.2f} a.u\n".format(symbol, cut))
            file.write(
                "----------------------------------------------------------------\n"
            )
        ## Write conduction cuts
        if conduction_cuts:
            file.write("Conduction correction cuts:\n")
            for key, value in conduction_cuts.items():
                symbol = key
                cut = value
                file.write("\t({}):{:.2f} a.u\n".format(symbol, cut))
            file.write(
                "----------------------------------------------------------------\n"
            )
        ## Write gap
        file.write("GAP: {:.3}eV".format(gap))



"""
List exchange and correlation codes for the INP file
"""
from enum import Enum, unique


@unique
class ExchangeCorreltion(Enum):
    """
    Enum type for exchange and correlation codes
    """

    ca = "ca"  # Ceperley-Alder
    wi = "wi"  # Wigner
    hl = "hl"  # Hedin-Lundqvist
    gl = "gl"  # Gunnarson-Lundqvist
    bh = "bh"  # Von Barth-Hedin
    pb = "pb"  # PBE scheme by Perdew, Burke, and Ernzerhof
    rp = "rp"  # RPBE scheme by Hammer, Hansen, and Norskov
    rv = "rv"  # revPBE scheme by Zhang and Yang
    bl = "bl"  # BLYP (Becke-Lee-Yang-Parr) scheme

    def __str__(self):
        return str(self.name)

    @staticmethod
    def to_list():
        """
        Generate list of exchange and correlation codes
        """
        return list(map(lambda element: element.value, ExchangeCorreltion))



"""
Makes fractional occupation on INP file
"""
import os
import sys
import click
import numpy as np
from loguru import logger
from minushalf.utils import (welcome_message, end_message, InputFile)
from minushalf import atomic_program


@click.command()
@click.argument('orbital_quantum_number', type=str, nargs=1)
@click.argument('occupation_percentage', type=str, nargs=1, default="100")
@click.option('--quiet', default=False, is_flag=True)
def occupation(orbital_quantum_number: str, occupation_percentage: str,
               quiet: bool):
    """
    Perform fractional occupation on the atom and generate the potential for this occupation.
    The occupation can subtract any fraction of the electron between 0 and 0.5, half occupation is the default.

    Requires:

        ORBITAL_QUANTUM_NUMBER: A string that defines the orbital(s) in which the occupation will be made,
        it can assume four values: (0: s | 1: p | 2: d | 3: f). if going to modify multiple orbitals,
        pass a string with numbers separated by commas : ("0,1,2,3").


        OCCUPATION_PERCENTAGE: A string that defines percentual of half an electron to be used in the occupation.
        The default is 100%, which states for 0.5e. For multiple occupations in different orbitals, pass a string
        separated by commas ("100,50,40,100"). For simplicity, to avoid the excessive repetition of the number
        100, just replace the number with * ("*,30,*"). If this argument is not used, the occupation of
        half an electron will be made for all orbitals assed as arguments.


        INP: Input file of the run-atomic command.

    Returns:

        INP_OCC : Input file modified for fractional occupation.


        INP.ae: A copy of the input file for the calculation.


        VTOTAL_OCC: Contains the atom potential for fractional occupation.


        OUT: Contains detailed information about the run.


        AECHARGE: Contains in four columns values of r, the “up” and “down” parts of the total
        charge density, and the total core charge density (the charges multiplied by 4πr 2 ).


        CHARGE: is exactly identical to AECHARGE and is generated for backwards compatibility.


        RHO: Like CHARGE, but without the 4πr 2 factor


        AEWFNR0...AEWFNR3: All-electron valence wavefunctions as function of radius, for s, p, d,
        and f valence orbitals (0, 1, 2, 3, respectively — some channels might not be available).
        They include a factor of r, the s orbitals also going to zero at the origin.
    """

    welcome_message("minushalf")

    if quiet:
        logger.remove()
        logger.add(sys.stdout, level="ERROR")

    input_file = InputFile.from_file()
    logger.info("Adding minus one half electron correction on INP")

    try:
        quantum_numbers = np.array(orbital_quantum_number.split(","),
                                   dtype=int)
    except ValueError as wrong_input:
        raise ValueError(
            "Invalid value for secondary quantum number") from wrong_input

    for quantum_number in quantum_numbers:
        if quantum_number < 0 or quantum_number > 3:
            raise ValueError("Invalid value for secondary quantum number")

    percentuals = np.ones(len(quantum_numbers)) * 100

    for index, percentual in enumerate(occupation_percentage.split(",")):
        if percentual != "*":
            if float(percentual) < 0 or float(percentual) > 100:
                raise ValueError("Invalid value for occupation percentual")
            percentuals[index] = float(percentual)

    for quantum_number, percentual in zip(quantum_numbers, percentuals):
        input_file.electron_occupation(0.5 * (percentual / 100),
                                       quantum_number)

    os.rename('INP', 'INP.ae')
    input_file.to_file()

    logger.info("Run atomic program")
    try:
        atomic_program.run()
    except Exception as program_fail:
        raise Exception('Problems in atomic program') from program_fail

    logger.info("Atomic program finished execution.")

    if not os.path.exists('./VTOTAL1'):
        raise FileNotFoundError("Problems in INP file generation")
    logger.info("Changing VTOTAL1 to VTOTAL_OCC")
    os.rename("VTOTAL1", "VTOTAL_OCC")

    logger.info("Changing INP to INP_OCC")
    os.rename('INP', 'INP_OCC')

    end_message()



"""
Analyze VTOTAL
"""
import re
from itertools import islice, chain
import numpy as np


class Vtotal():
    """
    Output for ATOM that contains the
    pseudopotential generated by an atom
    """
    def __init__(self, radius: np.array, down_potential: np.array) -> None:
        """
            Args:

                radius (np.array):Rays for which the potential calculations will be made made.

                down_potential (np.array):Potentials calculated to the state of spin Down for
                                          each value of radius
        """
        self.vtotal_header_size = 1
        self.radius = radius
        self.down_potential = down_potential

    @staticmethod
    def from_file(filename: str = "./VTOTAL.ae") -> any:
        """
        Parse VTOTAL and extract the following informations
        """
        radius = Vtotal.read_radius(filename)
        down_potential = Vtotal.read_down_potential(filename)
        return Vtotal(radius, down_potential)

    @staticmethod
    def read_down_potential(filename: str) -> np.array:
        """
        Extracts the potentials related to the state
        of spin Down calculated for the main elements

            Args:
                filename (str): Name of the VTOTAL file
        """

        with open(filename, "r") as vtotal:
            initial_regex = re.compile(r"^.*Down\s+potential\s+follows")
            for line in vtotal:
                if initial_regex.match(line):
                    break
            else:
                raise ValueError(
                    "Potential informations do not found in vtotal")

            stop_regex = re.compile(r"^.*Up\s+potential\s+follows")
            down_potential = []
            vtotal_header_size = 1  # Skip the line containing the spin value

            for line in islice(vtotal, vtotal_header_size, None):

                if stop_regex.match(line):
                    down_potential = list(chain.from_iterable(down_potential))
                    return np.array(down_potential, dtype=np.float64)

                down_potential.append(line.split())

    @staticmethod
    def read_radius(filename: str) -> np.array:
        """
        Extracts from the file information regarding the rays for
        which the potential calculations will be made made.

            Args:
                filename (str): Name of the VTOTAL file
        """

        with open(filename, "r") as vtotal:
            radius = []
            vtotal_header_size = 1  # Skip the header Raios
            stop_regex = re.compile(r"^.*Down\s+potential\s+follows")
            for line in islice(vtotal, vtotal_header_size, None):

                if stop_regex.match(line):
                    radius = list(chain.from_iterable(radius))
                    break

                radius.append(line.split())
        return np.array(radius, dtype=np.float64)



"""
Run atomic program
"""
import os
import sys
import click
from loguru import logger
from minushalf.utils import welcome_message, end_message
from minushalf import atomic_program


@click.command()
@click.option('--quiet', default=False, is_flag=True)
def run_atomic(quiet: bool):
    """
    Run the atomic program. The atomic program used is a version of the
    ATOM modified by the professor Luiz Guimarães Ferreira to be used in
    the DFT -1/2 method.

    Requires:

        INP: The input file for the calculation.

    Returns:

        VTOTAL.ae: Contains the atom potential.


        OUT: Contains detailed information about the run.


        AECHARGE: Contains in four columns values of r, the “up” and “down” parts of the total
                  charge density, and the total core charge density (the charges multiplied by 4πr 2 ).

        CHARGE: is exactly identical to AECHARGE and is generated for backwards compatibility.


        RHO: Like CHARGE, but without the 4πr 2 factor


        AEWFNR0...AEWFNR3: All-electron valence wavefunctions as function of radius, for s, p, d,
                           and f valence orbitals (0, 1, 2, 3, respectively — some channels might not be available).
                           They include a factor of r, the s orbitals also going to zero at the origin.
    """
    welcome_message("minushalf")

    if quiet:
        logger.remove()
        logger.add(sys.stdout, level="ERROR")

    logger.info("Run atomic program")

    try:
        atomic_program.run()
    except:
        raise Exception('Problems in atomic program execution')

    logger.info("Atomic program finished execution.")

    if not os.path.exists('./VTOTAL1'):
        logger.warning(
            "VTOTAL1 not found, thus VTOTAL.ae will not be generated")
    else:
        logger.info("Changing VTOTAL1 to VTOTAL.ae")
        os.rename("VTOTAL1", "VTOTAL.ae")

    end_message()



"""
Aims to show how the last conduction band and the
first valence band are composed by the orbitals of each atom.
"""
import click
from minushalf.data import Softwares
from minushalf.softwares import VaspFactory
from minushalf.utils import (welcome_message, end_message, projection_to_df,
                             BandStructure)


@click.command()
@click.option(
    '-s',
    '--software',
    type=click.Choice(Softwares.to_list(), case_sensitive=False),
    default=Softwares.vasp.value,
    show_default=True,
    help="""Specifies the software used to perform ab initio calculations.""")
@click.option('-b',
              '--base-path',
              type=click.Path(),
              nargs=1,
              help="""Path to folder where the relevant files are located.""")
def cbm_character(software: str, base_path: str) -> None:
    """Uses output files from softwares that perform ab initio calculations to
      discover the first conduction band (CBM) and extract, in percentage, its
      character corresponding to each orbital type (s, p, d, ... ). The
      names of the files required for each software are listed below, it is
      worth mentioning that their names cannot be modified.

    VASP: PROCAR, EIGENVAL, vasprun.xml
    """

    welcome_message("minushalf")

    softwares = {"VASP": VaspFactory()}

    factory = softwares[software.upper()]

    eigenvalues = factory.get_eigenvalues(base_path=base_path)
    fermi_energy = factory.get_fermi_energy(base_path=base_path)
    atoms_map = factory.get_atoms_map(base_path=base_path)
    num_bands = factory.get_number_of_bands(base_path=base_path)
    band_projection_file = factory.get_band_projection_class(
        base_path=base_path)

    band_structure = BandStructure(eigenvalues, fermi_energy, atoms_map,
                                   num_bands, band_projection_file)
    cbm_projection = band_structure.cbm_projection()
    normalized_df = projection_to_df(cbm_projection)
    click.echo(normalized_df.to_markdown())

    end_message()



"""
Reads Eigenval file, an output of
VASP software
"""
import re
from collections import defaultdict


class Eigenvalues():
    """
    Reads eigenvalues and store
    useful informations
    """
    def __init__(self, filename: str):
        """
            Args:
                filename (str): name of the EIGENVAL file in VASP

            Members:
                eigenvalues (defaultdict(list)): An dictionary where
                the keys are kpoint index and the values are lists
                with the eigenvalues for each band
        """
        self.filename = filename
        self.eigenvalues = self._get_eigenvalues()

    def _get_eigenvalues(self) -> defaultdict(list):
        """
            Returns:
                eigenvalues (defaultdict(list)): An dictionary where
                the keys are kpoint index and the values are lists
                with the eigenvalues for each band
        """
        eigenvalues = defaultdict(list)
        kpoint = 0
        band = None
        eigenvalue_regex = re.compile(r"\s*([0-9]+)\s+([-+]?[0-9]*\.[0-9]+)")
        with open(self.filename) as eigenval:
            for line in eigenval:
                if eigenvalue_regex.match(line):
                    band = int(eigenvalue_regex.match(line).group(1))
                    eigenvalue = float(eigenvalue_regex.match(line).group(2))

                    if band == 1:
                        kpoint += 1

                    eigenvalues[kpoint].append(eigenvalue)
        return eigenvalues



"""
Configuration setup for python project
"""
from os import path
import setuptools
from numpy.distutils.core import Extension, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

atomic_program = Extension(name="minushalf.atomic_program",
                           sources=[
                               "minushalf/atomic_program/atm_cGuima3.f",
                               "minushalf/atomic_program/atm_cGuima3.pyf",
                           ],
                           extra_compile_args=['-static'])

setup(
    name="minushalf",
    version="1.4",
    packages=setuptools.find_packages(),
    include_package_data=True,
    author="Henrique Fernandes",
    author_email="hentt30@gmail.com",
    description=
    "CLI to provides Pre processing tools for DFT -1/2 calculations",
    long_description=long_description,
    license="GPL",
    install_requires=[
        "numpy>=1.19.4",
        "pandas>=1.1.5",
        "fortranformat>=0.2.5",
        "Click>=7.1.2",
        "pyfiglet>=0.8",
        "loguru>=0.5.3",
        "tabulate>=0.8.7",
        "pyyaml",
        "scipy>=1.5.4",
        "aenum>=3.0.0",
    ],
    entry_points="""
        [console_scripts]
        minushalf=minushalf.minushalf:minushalf
    """,
    ext_modules=[atomic_program],
)



"""
Get atoms to be corrected for
simple valence correction and simple
condunction correction
"""
import itertools
import pandas as pd
from collections import defaultdict


def get_fractionary_correction_indexes(band_projection: pd.DataFrame,
                                        treshold: int = 5) -> dict:
    """
        Get dataframe index of the orbitals which contributes more
        than 5 percent to (VBM|CBM)

            Returns:

                correction_indexes (dict):A dict wherw the keys are the atoms
                                          symbols and the value is a list with the orbitals type to be
                                          corrected.
                                          Ex:
                                          {
                                          'Ga': ['p','s'],
                                          'N' : ['d','f'],
                                          }

    """
    columns_name = band_projection.columns.tolist()
    rows_name = band_projection.index.tolist()

    correction_indexes = defaultdict(list)
    for row, column in itertools.product(rows_name, columns_name):
        if band_projection[column][row] >= treshold:
            correction_indexes[row].append(column)
    return correction_indexes



"""
Physical constants
"""


class Constants:
    """
    Class for physical constants used in the program. Contains:

        pi: About 3.1415

        trimming exponent: exponent used in the trimming function

        bohr_radius: The Bohr radius is a physical constant, equal to
                     the most probable distance between the nucleus and
                     the electron in a hydrogen atom in its ground state.

        rydberg: In spectroscopy, the Rydberg constant, symbol for heavy
                 atoms or for hydrogen, named after the Swedish physicist
                 Johannes Rydberg, is a physical constant relating to the
                 electromagnetic spectra of an atom
    """
    @property
    def pi_constant(self):
        """
        Constant PI
        """
        return 3.1415926548

    @property
    def trimming_exponent(self):
        """
        Expoent for trimming function
        """
        return 8

    @property
    def bohr_radius(self):
        """
        Bohr radius
        """
        return 0.529177068

    @property
    def rydberg(self):
        """
        Rydberg constant
        """
        return 13.6058038



"""
    Factory to generate same modules for different softwares

    Supports the following softwares:
    - VASP
"""
import os
from typing import List
from minushalf.interfaces import SoftwaresAbstractFactory
from minushalf.utils import (
    check_procar_exists,
    check_vasprun_exists,
    check_eigenval_exists,
    check_potcar_exists,
    check_outcar_exists,
)
from minushalf.softwares.vasp import (
    Procar,
    Vasprun,
    Eigenvalues as VaspEigenval,
    Potcar,
    VaspRunner,
    Outcar,
)


class VaspFactory(SoftwaresAbstractFactory):
    """
    Concrete Factory for create instances
    for each supported software.
    """
    @check_vasprun_exists
    def get_atoms_map(self,
                      filename: str = "vasprun.xml",
                      base_path: str = None) -> dict:
        """
            Args:
                filename (str): Name of the vasprun.xml file.
                base_path (str): Path to the folder where the file is located.

            Returns:
                atoms_map (dict): Map of atomic symbols for their respective indexes
        """
        if base_path:
            filename = os.path.join(base_path, filename)
        vasprun = Vasprun(filename)
        return vasprun.atoms_map

    @check_vasprun_exists
    def get_fermi_energy(self,
                         filename: str = "vasprun.xml",
                         base_path: str = None) -> float:
        """
            Args:
                filename (str): Name of the vasprun.xml file.
                base_path (str): Path to the folder where the file is located.

            Returns:
                fermi_energy (dict): Energy of the fermi level
        """
        if base_path:
            filename = os.path.join(base_path, filename)
        vasprun = Vasprun(filename)
        return vasprun.fermi_energy

    @check_procar_exists
    def get_band_projection_class(
        self,
        filename: str = "PROCAR",
        base_path: str = None,
    ) -> Procar:
        """
            Args:
                filename (str): Name of the PROCAR file
                base_path (str): Path to the folder where the file is located.

            Returns:
                procar (Procar): Contains the class that handles files
                that contains informations about band projections
        """
        if base_path:
            filename = os.path.join(base_path, filename)
        return Procar(filename)

    @check_procar_exists
    def get_number_of_bands(self,
                            filename: str = "PROCAR",
                            base_path: str = None) -> int:
        """
            Args:
                filename (str): Name of the PROCAR file
                base_path (str): Path to the folder where the file is located.

            Returns:
                number_of_bands(int): Number of bands used in calculation
        """
        if base_path:
            filename = os.path.join(base_path, filename)
        procar = Procar(filename)
        return procar.num_bands

    @check_procar_exists
    def get_number_of_kpoints(self,
                              filename: str = "PROCAR",
                              base_path: str = None) -> int:
        """
            Args:
                filename (str): Name of the PROCAR file.
                base_path (str): Path to the folder where the file is located.

            Returns:
                number_of_kpoints(int): Number of kpoints used in calculation
        """
        if base_path:
            filename = os.path.join(base_path, filename)
        procar = Procar(filename)
        return procar.num_kpoints

    @check_potcar_exists
    def get_potential_class(
        self,
        filename: str = "POTCAR",
        base_path: str = None,
    ) -> Potcar:
        """
            Args:
                filename (str): Name of the POTCAR file.
                base_path (str): Path to the folder where the file is located.

            Returns:
                Potcar: class to the potential file
        """
        if base_path:
            filename = os.path.join(base_path, filename)
        return Potcar(filename)

    @check_eigenval_exists
    def get_eigenvalues(self,
                        filename: str = "EIGENVAL",
                        base_path: str = None) -> dict:
        """
            Args:
                filename (str): Name of the EIGENVAL file
                base_path (str): Path to the folder where the file is located.

            Returns:
                eigenvalues (dict): dictionary containing the eigenvalues
                for each kpoint and each band
        """
        if base_path:
            filename = os.path.join(base_path, filename)
        eigenval = VaspEigenval(filename)
        return eigenval.eigenvalues

    def get_runner(self, command: List[str]):
        """
        Return the class
        that runs VASP
        """
        return VaspRunner(command)

    @check_outcar_exists
    def get_nearest_neighbor_distance(self,
                                      ion_index: str,
                                      filename: str = "OUTCAR",
                                      base_path: str = None) -> float:
        """
            Args:
                ion_index (str): The index of the ion given by VASP.
                filename (str): Name of the OUTCAR file.
                base_path (str): Path to the folder where the file is located.

            Returns:
                distance (float): The distance of the nearest neighbor.
        """
        if base_path:
            filename = os.path.join(base_path, filename)

        outcar = Outcar(filename)
        return outcar.nearest_neighbor_distance(ion_index)

    @check_outcar_exists
    def get_number_of_equal_neighbors(self,
                                      atoms_map: dict,
                                      symbol: str,
                                      filename: str = "OUTCAR",
                                      base_path: str = None) -> float:
        """
        Given an map that links atoms symbols with it's index
        this function returns the number of neighbors of the atom with
        equal symbol but different indexes.

            Args:
                atoms_map (dict): Map the atoms index to their symbol.
                symbom (str): The symbol of the target atom.

            Returns:
                number_equal_neighbors (int): Returns the number of neighbors with
                                        same symbol but different indexes.
        """
        if base_path:
            filename = os.path.join(base_path, filename)

        outcar = Outcar(filename)
        return outcar.number_of_equal_neighbors(atoms_map, symbol)



"""
Commands of the minhushalf cli
"""
from .correct_potfile import correct_potfile
from .run_atomic_program import run_atomic
from .fractional_occupation import occupation
from .vbm_character import vbm_character
from .cbm_character import cbm_character
from .band_character import band_character
from .create_input import create_input
from .band_gap import band_gap
from .execute import execute



"""
Reads vasprun.xml file, an output of
VASP software
"""
import re
import xml.etree.ElementTree as ET


class Vasprun():
    """
	Reads vasprun.xml and store
	useful informations
	"""
    def __init__(self, filename: str):
        """
			Args:
				filename (str): name of the vasprun file in VASP
			Members:
				num_kpoints: Number of kpoints used in the simulation
				num_bands: Number of bands used in the simulation
				size_procar_header: Number of lines of the procar header in PROCAR file
		"""
        self.filename = filename
        self.fermi_energy = self._get_fermi_energy()
        self.atoms_map = self._get_atoms()

    def _get_fermi_energy(self) -> float:
        """
		Extract fermi energy from vasprun.xml file
			Returns:
				fermi_energy (float): Fermi energy
		"""
        efermi_xml = self._catch_xml_tag(tag="i", name="efermi")
        xml_tree = ET.fromstringlist(efermi_xml)
        try:
            fermi_energy = float(xml_tree.text)
        except ValueError as invalid_conversion:
            raise Exception('Vasprun parser does not find fermi energy'
                            ) from invalid_conversion
        finally:
            xml_tree.clear()
        return fermi_energy

    def _get_atoms(self) -> dict:
        """
		Extract informations about the
		atoms from vasprun.xml
			Returns:
				A dict containing the index of the atoms
				and their symbols. dict[index] =  symbol
		"""
        atoms_xml = self._catch_xml_tag(tag="array", name="atoms")

        atoms_tree = ET.fromstringlist(atoms_xml)

        try:
            atoms = {}
            for index, value in enumerate(
                    atoms_tree.find("set").findall("rc")):
                atom_index = list(value.findall("c"))
                atoms[str(index + 1)] = atom_index[0].text.strip()
        except:
            raise Exception("Vasprun parser do not found atoms informations")
        finally:
            atoms_tree.clear()
        return atoms

    def _catch_xml_tag(self, tag: str, name: str = None) -> list:
        """
		Catch element by the tag in a large xml file

		Args:
			tag (str): Tag to the element in xml

		Returns:
			captured_records (np.array): lines between the
			start tag and end tag
		"""

        start_tag_regex = re.compile(rf'^.*<{tag}\s*(name="{name}")?\s*>')
        end_tag_regex = re.compile(rf"^.*</{tag}\s*>")
        xml_text = []
        with open(self.filename, "r") as vasprun:
            start_tag_identified = False
            for line in vasprun:
                if start_tag_identified and end_tag_regex.match(line):
                    end_tag = end_tag_regex.search(line).group(0)
                    end_line_splited = line.partition(end_tag)
                    end_line = "".join(end_line_splited[:2])
                    xml_text.append(end_line)
                    break

                elif start_tag_identified:
                    xml_text.append(line)

                elif not start_tag_identified and start_tag_regex.match(line):
                    start_tag = start_tag_regex.search(line).group(0)
                    start_line_splited = line.partition(start_tag)
                    start_line = "".join(start_line_splited[1:])
                    start_tag_identified = True
                    xml_text.append(start_line)
                    if end_tag_regex.match(line):
                        break
        return xml_text
